#! /usr/bin/env python
############################################################################
#  notify_utils.py
#
#  Copyright 2020 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose: Provide a set of tools to notify users (e.g., of process
#          finishing).
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 02/04/2020
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
###########################################################################


def send_email_notification(message, subject, receiver_email, sender_email=None, smtp_server=None, smtp_port=None,
                            smtp_password=None):
    """
    A function which sends an email.

    :param message: a string with the message to be sent.
    :param subject: a string with the subject of your email.
    :param receiver_email: the email address(es) of the receiver. If you want to email more than one person
                           then separate the email addresses with commas.
    :param sender_email: The email address of the sender. If None then will read variable from RSGIS_SENDER_EMAIL.
    :param smtp_server: The SMTP server to send the email. If None then will read variable from RSGIS_SMPT_SERVER.
    :param smtp_port: The SMTP server port. If None then will read variable from RSGIS_SMTP_PORT.
    :param smtp_password: The password for the SMTP server. If None then will read variable from RSGIS_PASSWORD.

    """
    import smtplib
    import ssl
    from email.mime.text import MIMEText
    from email.mime.multipart import MIMEMultipart
    import rsgislib
    rsgis_utils = rsgislib.RSGISPyUtils()

    if sender_email is None:
        sender_email = rsgis_utils.getEnvironmentVariable("RSGIS_SENDER_EMAIL")
        if sender_email is None:
            raise Exception("No senders email has been provided")

    if smtp_server is None:
        smtp_server = rsgis_utils.getEnvironmentVariable("RSGIS_SMPT_SERVER")
        if smtp_server is None:
            raise Exception("No SMTP server has been provided")

    if smtp_port is None:
        smtp_port = rsgis_utils.getEnvironmentVariable("RSGIS_SMTP_PORT")
        if smtp_port is None:
            raise Exception("No SMTP port has been provided")

    if smtp_password is None:
        smtp_password = rsgis_utils.getEnvironmentVariable("RSGIS_PASSWORD")
        if smtp_password is None:
            raise Exception("No SMTP password has been provided")

    email_message = MIMEMultipart()
    email_message["Subject"] = subject
    email_message["From"] = sender_email
    email_message["To"] = receiver_email
    email_message.attach(MIMEText(message))
    print(email_message.as_string())

    try:
        context = ssl.create_default_context()
        with smtplib.SMTP_SSL(smtp_server, smtp_port, context=context) as server:
            server.login(sender_email, smtp_password)
            server.sendmail(sender_email, receiver_email, email_message.as_string())
        print("Email has been send to {}".format(receiver_email))
    except smtplib.SMTPException as e:
        print("Error failed to send email: '{}'".format(e))



#! /usr/bin/env python
############################################################################
#  classkeraspxl.py
#
#  Copyright 2020 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide a set of functions to apply tensorflow keras classifiers
#           to remotely sensed imagery.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 14/01/2020
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
###########################################################################

from __future__ import print_function

import rsgislib
import rsgislib.rastergis

import numpy

import h5py

import osgeo.gdal as gdal

from rios import applier
from rios import cuiprogress
from rios import rat


def train_keras_pixel_classifer(cls_mdl, clsinfodict, out_mdl_file=None, train_epochs=5, train_batch_size=32):
    """
    A function which trains a neural network defined using the keras API for the classification of remotely sensed data.
    A dict of class information, as ClassInfoObj objects, is defined with the training data.

    This function requires that tensorflow and keras modules to be installed.

    :param out_mdl_file: The output model which can be loaded to perform a classification.
    :param clsinfodict: dict (key is string with class name) of ClassInfoObj objects defining the training data.
    :param out_mdl_file: A file path to save the trained model as a hdf5 file. If None then ignored.

    """
    try:
        from keras.utils import to_categorical
    except:
        from tensorflow.keras.utils import to_categorical

    n_classes = len(clsinfodict)
    for clsname in clsinfodict:
        if clsinfodict[clsname].id >= n_classes:
            raise ("ClassInfoObj '{}' id ({}) is not consecutive starting from 0.".format(clsname,
                                                                                          clsinfodict[clsname].id))

    cls_data_dict = {}
    train_data_lst = []
    train_lbls_lst = []
    valid_data_lst = []
    valid_lbls_lst = []
    test_data_lst = []
    test_lbls_lst = []
    cls_ids = []
    for clsname in clsinfodict:
        sgl_cls_info = {}
        print("Reading Class {} Training".format(clsname))
        f = h5py.File(clsinfodict[clsname].trainfileH5, 'r')
        sgl_cls_info['train_n_rows'] = f['DATA/DATA'].shape[0]
        sgl_cls_info['train_data'] = numpy.array(f['DATA/DATA'])
        sgl_cls_info['train_data_lbls'] = numpy.zeros(sgl_cls_info['train_n_rows'], dtype=int)
        sgl_cls_info['train_data_lbls'][...] = clsinfodict[clsname].id
        f.close()
        train_data_lst.append(sgl_cls_info['train_data'])
        train_lbls_lst.append(sgl_cls_info['train_data_lbls'])

        print("Reading Class {} Validation".format(clsname))
        f = h5py.File(clsinfodict[clsname].validfileH5, 'r')
        sgl_cls_info['valid_n_rows'] = f['DATA/DATA'].shape[0]
        sgl_cls_info['valid_data'] = numpy.array(f['DATA/DATA'])
        sgl_cls_info['valid_data_lbls'] = numpy.zeros(sgl_cls_info['valid_n_rows'], dtype=int)
        sgl_cls_info['valid_data_lbls'][...] = clsinfodict[clsname].id
        f.close()
        valid_data_lst.append(sgl_cls_info['valid_data'])
        valid_lbls_lst.append(sgl_cls_info['valid_data_lbls'])

        print("Reading Class {} Testing".format(clsname))
        f = h5py.File(clsinfodict[clsname].testfileH5, 'r')
        sgl_cls_info['test_n_rows'] = f['DATA/DATA'].shape[0]
        sgl_cls_info['test_data'] = numpy.array(f['DATA/DATA'])
        sgl_cls_info['test_data_lbls'] = numpy.zeros(sgl_cls_info['test_n_rows'], dtype=int)
        sgl_cls_info['test_data_lbls'][...] = clsinfodict[clsname].id
        f.close()
        test_data_lst.append(sgl_cls_info['test_data'])
        test_lbls_lst.append(sgl_cls_info['test_data_lbls'])

        cls_data_dict[clsname] = sgl_cls_info
        cls_ids.append(clsinfodict[clsname].id)

    print("Finished Reading Data")
    
    train_np = numpy.concatenate(train_data_lst)
    train_lbls_np = numpy.concatenate(train_lbls_lst)
    train_lbls_keras = to_categorical(train_lbls_np, num_classes=n_classes)
    
    vaild_np = numpy.concatenate(valid_data_lst)
    vaild_lbls_np = numpy.concatenate(valid_lbls_lst)
    vaild_lbls_keras = to_categorical(vaild_lbls_np, num_classes=n_classes)

    test_np = numpy.concatenate(test_data_lst)
    test_lbls_np = numpy.concatenate(test_lbls_lst)
    test_lbls_keras = to_categorical(test_lbls_np, num_classes=n_classes)
    
    print("Start Training Model")
    cls_mdl.fit(train_np, train_lbls_keras, epochs=train_epochs, batch_size=train_batch_size,
                validation_data=(vaild_np, vaild_lbls_keras))
    print("Finished Training Model")
    cls_mdl.summary()
    
    loss_and_metrics = cls_mdl.evaluate(test_np, test_lbls_keras, batch_size=train_batch_size)
    eval_metric_names = cls_mdl.metrics_names
    for eval_name,eval_val in zip(eval_metric_names, loss_and_metrics):
        print("{} = {}".format(eval_name, eval_val))
    
    if out_mdl_file is not None:
        cls_mdl.save(out_mdl_file)


def apply_keras_pixel_classifier(classTrainInfo, keras_cls_mdl, imgMask, imgMaskVal, imgFileInfo, outClassImg,
                                 gdalformat, pred_batch_size=32, classClrNames=True):
    """
This function applies a trained single pixel keras model to an image. The function train_keras_pixel_classifer
can be used to train such as model. The output image will contain the hard membership of the predicted class. 

:param classTrainInfo: dict (where the key is the class name) of rsgislib.classification.ClassInfoObj
                       objects which will be used to train the classifier (i.e., train_keras_pixel_classifer()),
                       provide pixel value id and RGB class values.
:param keras_cls_mdl: a trained keras model object, with a input dimensions equivlent to the number of image
                      bands specified in the imgFileInfo input and output layer which provides an output array
                      of the length of the number of classes.
:param imgMask: is an image file providing a mask to specify where should be classified. Simplest mask is all the
                valid data regions (rsgislib.imageutils.genValidMask)
:param imgMaskVal: the pixel value within the imgMask to limit the region to which the classification is applied.
                   Can be used to create a heirachical classification.
:param imgFileInfo: a list of rsgislib.imageutils.ImageBandInfo objects (also used within
                    rsgislib.imageutils.extractZoneImageBandValues2HDF) to identify which images and bands are to
                    be used for the classification so it adheres to the training data.
:param outClassImg: Output image which will contain the hard classification.
:param gdalformat: is the output image format - all GDAL supported formats are supported.
:param pred_batch_size: the batch size used for the classification.
:param classClrNames: default is True and therefore a colour table will the colours specified in ClassInfoObj
                      and a ClassName (from classTrainInfo) column will be added to the output file.

    """

    def _applyKerasPxlClassifier(info, inputs, outputs, otherargs):
        outClassIdVals = numpy.zeros_like(inputs.imageMask, dtype=numpy.uint16)
        if numpy.any(inputs.imageMask == otherargs.mskVal):
            n_pxls = inputs.imageMask.shape[1] * inputs.imageMask.shape[2]
            outClassIdVals = outClassIdVals.flatten()
            imgMaskVals = inputs.imageMask.flatten()
            classVars = numpy.zeros((n_pxls, otherargs.numClassVars), dtype=numpy.float)
            # Array index which can be used to populate the output array following masking etc.
            ID = numpy.arange(imgMaskVals.shape[0])
            classVarsIdx = 0
            for imgFile in otherargs.imgFileInfo:
                imgArr = inputs.__dict__[imgFile.name]
                for band in imgFile.bands:
                    classVars[..., classVarsIdx] = imgArr[(band - 1)].flatten()
                    classVarsIdx = classVarsIdx + 1
            classVars = classVars[imgMaskVals == otherargs.mskVal]
            ID = ID[imgMaskVals == otherargs.mskVal]
            preds_idxs = numpy.argmax(otherargs.classifier.predict(classVars,
                                                                   batch_size=otherargs.pred_batch_size), axis=1)
            preds_cls_ids = numpy.zeros_like(preds_idxs, dtype=numpy.uint16)
            for cld_id, idx in zip(otherargs.cls_id_lut, numpy.arange(0, len(otherargs.cls_id_lut))):
                preds_cls_ids[preds_idxs == idx] = cld_id

            outClassIdVals[ID] = preds_cls_ids
            outClassIdVals = numpy.expand_dims(
                outClassIdVals.reshape((inputs.imageMask.shape[1], inputs.imageMask.shape[2])), axis=0)
        outputs.outclsimage = outClassIdVals

    infiles = applier.FilenameAssociations()
    infiles.imageMask = imgMask
    numClassVars = 0
    for imgFile in imgFileInfo:
        infiles.__dict__[imgFile.name] = imgFile.fileName
        numClassVars = numClassVars + len(imgFile.bands)

    n_classes = len(classTrainInfo)
    cls_id_lut = numpy.zeros(n_classes)
    for clsname in classTrainInfo:
        if classTrainInfo[clsname].id >= n_classes:
            raise ("ClassInfoObj '{}' id ({}) is not consecutive starting from 0.".format(clsname,
                                                                                          classTrainInfo[clsname].id))
        cls_id_lut[classTrainInfo[clsname].id] = classTrainInfo[clsname].out_id

    outfiles = applier.FilenameAssociations()
    outfiles.outclsimage = outClassImg
    otherargs = applier.OtherInputs()
    otherargs.classifier = keras_cls_mdl
    otherargs.pred_batch_size = pred_batch_size
    otherargs.mskVal = imgMaskVal
    otherargs.numClassVars = numClassVars
    otherargs.imgFileInfo = imgFileInfo
    otherargs.n_classes = n_classes
    otherargs.cls_id_lut = cls_id_lut

    try:
        import tqdm
        progress_bar = rsgislib.TQDMProgressBar()
    except:
        progress_bar = cuiprogress.GDALProgressBar()

    aControls = applier.ApplierControls()
    aControls.progress = progress_bar
    aControls.drivername = gdalformat
    aControls.omitPyramids = True
    aControls.calcStats = False
    print("Applying the Classifier")
    applier.apply(_applyKerasPxlClassifier, infiles, outfiles, otherargs, controls=aControls)
    print("Completed Classification")

    if classClrNames:
        rsgislib.rastergis.populateStats(outClassImg, addclrtab=True, calcpyramids=True, ignorezero=True)
        ratDataset = gdal.Open(outClassImg, gdal.GA_Update)
        red = rat.readColumn(ratDataset, 'Red')
        green = rat.readColumn(ratDataset, 'Green')
        blue = rat.readColumn(ratDataset, 'Blue')
        ClassName = numpy.empty_like(red, dtype=numpy.dtype('a255'))
        ClassName[...] = ""

        for classKey in classTrainInfo:
            print("Apply Colour to class \'" + classKey + "\'")
            red[classTrainInfo[classKey].out_id] = classTrainInfo[classKey].red
            green[classTrainInfo[classKey].out_id] = classTrainInfo[classKey].green
            blue[classTrainInfo[classKey].out_id] = classTrainInfo[classKey].blue
            ClassName[classTrainInfo[classKey].out_id] = classKey

        rat.writeColumn(ratDataset, "Red", red)
        rat.writeColumn(ratDataset, "Green", green)
        rat.writeColumn(ratDataset, "Blue", blue)
        rat.writeColumn(ratDataset, "ClassName", ClassName)
        ratDataset = None




#!/usr/bin/env python
"""
The tools.stats module contains some useful tools for calculating useful statistics which aren't
easily available else where.
"""

def calculate_VIF(X):
    """
    A function to calculate variance inflation factors to
    investigate multicollinearity between predictor variables. 
    Function written by Osian Roberts.

    :param X: a 2D numpy array of shape = (nSamples, nPredictors).
    :returns: VIF_scores; A numpy array containing the VIF for each predictor variable.

    Interpretation of VIF scores (this is somewhat subjective):
    1 = No multicollinearity.
    1-5 = Moderate multicollinearity.
    > 5 = High multicollinearity.
    > 10 = This predictor should be removed from the model.
    
    Example::
    
    import numpy
    # create some fake predictors to test the function:
    x1 = numpy.random.uniform(1, 100, 500)  # this should be highly collinear with x3.
    x2 = numpy.random.uniform(1, 100, 500)  # this should be independent of x1.
    x3 = 0.85 * x1 + 50 # this should be highly collinear with x1.
    x4 = 0.85 * numpy.log10(x1) + 50 # this should be moderately collinear with x1.
    
    X = numpy.array([x1, x2, x3, x4]).T
    
    vif_scores = calculate_VIF(X)
        
    for i in range(vif_scores.size):
        print('VIF for x' + str(i + 1) + ':', vif_scores[i])
    
    """
    import numpy
    from sklearn.linear_model import LinearRegression
    
    print('Calculating VIF for {} predictors variables...'.format(X.shape[1]))
    # fit auxiliary regressions to each predictor in X:
    lm = LinearRegression()
    VIF_scores = []
    for i in range(X.shape[1]):
        # fit an OLS linear model to each predictor (y) using the remaining predictor variables (x):
        y = X[:, i]
        x = numpy.delete(X, i, axis=1)
        lm.fit(x, y)
        y_pred = lm.predict(x)
        del x

        # calculate the coefficient of determination:
        ss_tot = sum((y - numpy.mean(y))**2)
        ss_res = sum((y - y_pred)**2)
        r2 = 1 - (ss_res / ss_tot)
        del y, y_pred, ss_res, ss_tot

        # calculate VIF:
        if r2 == 1:
            r2 = 0.999  # this prevents division by zero errors when r2 == 1.
        VIF_scores.append(1 / (1 - r2))
        del r2
    del lm, X
    VIF_scores = numpy.array(VIF_scores).round(decimals=1)
    return VIF_scores


def bin_accuracy_scores_prob(y_true, y_prob):
    """
    A function to calculate accuracy measures for probabilistic responses with sklearn and scipy.
    Function written by Osian Roberts.

    :param y_true: binary class labels, where 0 is absence and 1 is presence.
    :param y_prob: probability of presence scores e.g., generated by a species distribution model.

    :returns: a list containing two arrays - metrics = names of test metrics. scores = test scores for each metric.

    Useful reference:
    https://machinelearningmastery.com/how-to-score-probability-predictions-in-python
    """
    import numpy
    # check inputs:
    if not isinstance(y_true, numpy.ndarray):
        y_true = numpy.array(y_true)
    if not isinstance(y_prob, numpy.ndarray):
        y_prob = numpy.array(y_prob)
    if y_true.ndim != 1:
        raise SystemExit('ERROR: the true labels are not in a 1D array.')
    if y_prob.ndim != 1:
        raise SystemExit('ERROR: the probability of presence values are not in a 1D array.')
    if y_true.size != y_prob.size:
        raise SystemExit('ERROR: unequal number of binary labels and probabilities.')

        # ensure that y_true contains binary labels (i.e. 0 or 1 values):
    y_true = y_true.astype('uint8')
    if numpy.min(y_true) != 0 or numpy.max(y_true) != 1:
        raise SystemExit('ERROR: the true labels are not binary (zero or one values).')

    from sklearn.metrics import roc_auc_score
    # calculates area under the receiver operating curve score.
    # A score of 0.5 shows the model is unable to discriminate between presence and absence.
    roc_auc = roc_auc_score(y_true, y_prob)

    from sklearn.metrics import average_precision_score
    # calculates area under the precision-recall curve. Perfect model = 1.0.
    average_precision = average_precision_score(y_true, y_prob)

    from sklearn.metrics import brier_score_loss
    # This is a quadratic loss function that calculates the mean squared error between
    # predicted probabilities and the true presence-absence (binary) labels.
    # A model with no false positives/negatives has a score of 0.0. Perfect model = 1.0.
    brier_score = brier_score_loss(y_true, y_prob)

    from sklearn.metrics import log_loss
    # The is logarithmic loss function that more heavily penalises false positives/negatives than the brier score.
    # A model with no false positives/negatives has a score of 0.0. There is no upper bound.
    log_loss_score = log_loss(y_true, y_prob)

    from scipy.stats import pointbiserialr
    # The point biserial correlation coefficient, range -1 to 1.
    # Quantifies the correlation between a binary and continuous variable.
    r = pointbiserialr(y_true, y_prob)[0]

    metrics = ['Test AUC', 'Point-Biserial r', 'Av. Precision', 'Brier Score', 'Log-Loss Score']
    scores = numpy.array([roc_auc, r, average_precision, brier_score, log_loss_score]).round(decimals=6)
    del roc_auc, r, average_precision, brier_score, log_loss_score, y_true, y_prob
    return [metrics, scores]


def accuracy_scores_binary(y_true, y_pred):
    '''
    A function to calculate accuracy measures for a binary classification.
    Function written by Osian Roberts.

    Parameters:
    :param y_true: observed binary labels, where 0 is absence and 1 is presence.
    :param y_pred: predicted binary labels, where 0 is absence and 1 is presence.

    :returns: a list containing two numpy.arrays - (metrics: name of test metrics, scores: test scores for each metric)

    Reference: See pages 253 - 255 in:
    Guisan et al. (2017). Habitat suitability and distribution models: with applications in R.
    '''
    import numpy

    # check inputs:
    if not isinstance(y_true, numpy.ndarray):
        y_true = numpy.array(y_true)
    if not isinstance(y_pred, numpy.ndarray):
        y_pred = numpy.array(y_pred)
    if y_true.ndim != 1:
        raise SystemExit('ERROR: the true labels are not in a 1D array.')
    if y_pred.ndim != 1:
        raise SystemExit('ERROR: the predicted labels are not in a 1D array.')
    if y_true.size != y_pred.size:
        raise SystemExit('ERROR: unequal number of binary labels.')

    # ensure that y_true, y_pred contain binary labels (i.e. 0 or 1 values):
    y_true = y_true.astype('uint8')
    y_pred = y_pred.astype('uint8')
    if numpy.min(y_true) != 0 or numpy.max(y_true) != 1:
        raise SystemExit('ERROR: the true labels are not binary (zero or one values).')
    if numpy.min(y_pred) != 0 or numpy.max(y_pred) != 1:
        raise SystemExit('ERROR: the predicted labels are not binary (zero or one values).')

    metrics = numpy.array(['Prevalence', 'Overall Diagnostic Power', 'Correct Classification Rate',
                           'Misclassification Rate', 'Presence Predictive Power', 'Absence Predictive Power',
                           'Accuracy', 'Balanced Accuracy', 'Sensitivity', 'Specificity', 'Precision', 'F1 Score',
                           'Matthews Correlation', 'Cohen Kappa', 'Normalised Mutual Information',
                           'Hanssen-Kuiper skill'])

    try:
        n_presence = numpy.where(y_true == 1)[0].size
        n_absence = numpy.where(y_true == 0)[0].size

        # calculate true-presence, true-absence, false-presence and false-absence:
        TP = numpy.where((y_true == 1) & (y_pred == 1))[0].size
        TA = numpy.where((y_true == 0) & (y_pred == 0))[0].size
        FP = numpy.where((y_true == 1) & (y_pred == 0))[0].size
        FA = numpy.where((y_true == 0) & (y_pred == 1))[0].size  # aka sweet FA!

        # proportion of presence records:
        prevalence = (TP / FA) / y_true.size

        # proportion of absence records:
        ODP = 1 - prevalence

        # correct classification & misclassification rate
        CCR = (TP + TA) / y_true.size
        MR = (FP + FA) / y_true.size

        # Sensitivity (aka Recall or True Positive Rate):
        sensitivity = TP / n_presence

        # false presence rate - inverse of sensitivity (redundant?)
        #FPR = 1  - sensitivity

        # Presence and absence predictive power:
        PPP = TP / (TP + FP)
        APP = TA / (TA + FA)

        # Specificity (aka True Negative Rate):
        specificity = TA / n_absence

        # false positive rate - inverse of specificity (redundant?)
        #FPR = 1 - specificity

        # Accuracy scores:
        accuracy = (TP + TA) / (n_presence + n_absence)
        balanced_accuracy = ((TP / n_presence) + (TA / n_absence)) / 2

        # precision:
        precision = TP / (TP + FP)

        # F1 score:
        f1_score = 2 * TP / ((2*TP) + FP + FA)

        # Matthews Correlation Coefficient:
        MCC = ((TP * TA) - (FP * FA)) / (((TP + FP) * (TP + FA) * (TA + FP) * (TA + FA))**0.5)

        # Hanssen-Kuiper skill (unreliable when TA is very large):
        TSS = sensitivity + specificity - 1
        del n_presence, n_absence, TP, TA, FP, FA

        from sklearn.metrics import normalized_mutual_info_score as nmi_score
        nmi_score = nmi_score(y_true, y_pred)

        # Cohen's Kappa (caution: sensitive to sample size and proportion of presence records):
        from sklearn.metrics import cohen_kappa_score as kappa
        kappa = kappa(y_true, y_pred)

        scores = numpy.array([prevalence, ODP, CCR, MR, PPP, APP, accuracy, balanced_accuracy, sensitivity,
                           specificity, precision, f1_score, MCC, kappa, nmi_score, TSS]).round(decimals=6)
        del prevalence, ODP, CCR, MR, PPP, APP, accuracy, balanced_accuracy, sensitivity
        del specificity, precision, f1_score, MCC, kappa, nmi_score, TSS
    except Exception:
        scores = numpy.zeros(len(metrics))

    if metrics.size == scores.size:
        return [metrics, scores]
    else:
        raise SystemExit('ERROR: unable to calculate accuracy metrics.')




#! /usr/bin/env python
############################################################################
#  classaccuracymetrics.py
#
#  Copyright 2020 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide a set of functions to calculate the accuracy of a
#           classification.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 03/02/2020
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
###########################################################################

import numpy


def cls_quantity_accuracy(y_true, y_pred, cls_area):
    """
    A function to calculate quantity allocation & disagreement for a
    land cover classification. The labels must be integers from 1 - N,
    where N is the number of classes.

    :param y_true: A list or 1D numpy array of true labels.
    :param y_pred: A list or 1D numpy array of predicted labels.
    :param cls_area: A dict or 1D numpy array of area/n_pixels identified by the
                     classifier. len(cls_area) == numpy.unique(y_true).

    :return: dict with 'Quantity Disagreement (Q)',
                       'Allocation Disagreement (A)',
                       'Proportion Correct (C)',
                       'Total Disagreement (D)'.

    Reference: Pontius, R. G., Jr, & Millones, M. (2011). Death to Kappa: birth
    of quantity disagreement and allocation disagreement for accuracy assessment.
    International Journal of Remote Sensing, 32(15), 4407–4429.

    """
    from sklearn.metrics import confusion_matrix

    # check inputs:
    if not isinstance(y_true, numpy.ndarray):
        y_true = numpy.array(y_true)
    if not isinstance(y_pred, numpy.ndarray):
        y_pred = numpy.array(y_pred)
    if not isinstance(cls_area, numpy.ndarray):
        cls_area = numpy.array(cls_area)

    for arr in [y_true, y_pred, cls_area]:
        if arr.ndim != 1:
            raise SystemExit('Error: All input arrays must be one dimensional.')

    if numpy.unique(y_true).size != cls_area.size:
        raise SystemExit('Error: Number of classes != Number of classes in area.')

    # create confusion matrix:
    cm = confusion_matrix(y_true, y_pred)

    # convert absolute areas into proportional areas:
    prop_area = (cls_area / cls_area.sum()).reshape(-1, 1)  # same as Comparison Total (see Ref.)

    # normalise the confusion matrix by proportional area:
    norm_cm = cm.astype(float) / cm.sum(axis=1)[:, ].reshape(-1, 1)
    norm_cm = norm_cm * prop_area
    comp_total = norm_cm.sum(axis=1)  # same as proportional area
    ref_total = norm_cm.sum(axis=0)

    quantity_disagreement = sum(numpy.abs(ref_total - comp_total)) / 2
    commission = [(row.sum() - row[idx]) for idx, row in enumerate(norm_cm)]
    ommission = ref_total - numpy.diag(norm_cm)
    allocation_disagreement = sum(2 * numpy.min(numpy.array([commission, ommission]), axis=0)) / 2
    prop_correct = sum(numpy.diag(norm_cm)) / numpy.sum(norm_cm)
    disagreement = quantity_disagreement + allocation_disagreement

    out_dict = dict()
    out_dict['Quantity Disagreement (Q)'] = quantity_disagreement
    out_dict['Allocation Disagreement (A)'] = allocation_disagreement
    out_dict['Proportion Correct (C)'] = prop_correct
    out_dict['Total Disagreement (D)'] = disagreement

    return out_dict


def calc_class_accuracy_metrics(ref_samples, pred_samples, cls_area=None, cls_names=None):
    """
    A function which calculates a set of classification accuracy metrics for a set
    of reference and predicted samples. Optionally, the area classified for each
    class can be provided allowing further metrics to be calculated.

    :param ref_samples: a 1d array of reference samples represented by a numeric class id
    :param pred_samples: a 1d array of predicted samples represented by a numeric class id
    :param cls_area: a 1d array with the area of each class classified (i.e., pixel count)
    :param cls_names: a 1d list of the class names (labels) in the order of the class ids.

    """
    import sklearn.metrics

    acc_metrics = sklearn.metrics.classification_report(ref_samples, pred_samples, target_names=cls_names,
                                                        output_dict=True)
    cohen_kappa = sklearn.metrics.cohen_kappa_score(ref_samples, pred_samples)
    acc_metrics['cohen_kappa'] = cohen_kappa

    cm = sklearn.metrics.confusion_matrix(ref_samples, pred_samples)
    user_accuracy = [(row[idx] / row.sum()) * 100 for idx, row in enumerate(cm)]
    producer_accuracy = [(col[idx] / col.sum()) * 100 for idx, col in enumerate(cm.T)]

    # convert absolute areas into proportional areas:
    prop_area = (cls_area / cls_area.sum()).reshape(-1, 1)  # same as Comparison Total (see Ref.)
    # normalise the confusion matrix by proportional area:
    norm_cm = cm.astype(float) / cm.sum(axis=1)[:, ].reshape(-1, 1)
    norm_cm = norm_cm * prop_area
    comp_total = norm_cm.sum(axis=1)  # same as proportional area
    ref_total = norm_cm.sum(axis=0)
    commission = [(row.sum() - row[idx]) for idx, row in enumerate(norm_cm)]
    ommission = ref_total - numpy.diag(norm_cm)
    # Sum the normalised cm columns to estimate the proportion of scene for each class.
    cls_area_prop = numpy.sum(norm_cm, axis=0)

    acc_metrics['confusion_matrix'] = cm.tolist()
    acc_metrics['user_accuracy'] = user_accuracy
    acc_metrics['producer_accuracy'] = producer_accuracy

    acc_metrics['norm_confusion_matrix'] = norm_cm.tolist()
    acc_metrics['commission'] = commission
    acc_metrics['ommission'] = ommission.tolist()
    acc_metrics['est_prop_cls_area'] = cls_area_prop.tolist()

    if cls_area is not None:
        quantity_metrics = cls_quantity_accuracy(ref_samples, pred_samples, cls_area)
        acc_metrics['quantity_metrics'] = quantity_metrics

    return acc_metrics


def calc_acc_metrics_vecsamples(in_vec_file, in_vec_lyr, ref_col, cls_col, cls_img, img_cls_name_col='ClassName',
                                img_hist_col='Histogram', out_json_file=None, out_csv_file=None):
    """
    A function which calculates classification accuracy metrics using a set of
    reference samples in a vector file.
    This would be often be used alongside the ClassAccuracy QGIS plugin.

    :param in_vec_file: the input vector file with the reference points
    :param in_vec_lyr: the input vector layer name with the reference points.
    :param ref_col: the name of the reference classification column in the input vector file.
    :param cls_col: the name of the classification column in the input vector file.
    :param cls_img: an image of the classification from which the area
                    (pixel counts) of each class are extracted to normalise the
                    confusion matrix. Should have a RAT with class names and histogram.
    :param img_cls_name_col: The name of the column in the image attribute table which specifies the
                             class name.
    :param img_hist_col: The name of the column in the image attribute table which contains the
                         histogram (i.e., number of pixels within the class).
    :param out_json_file: if specified the generated metrics and confusion matrix are written to
                          a JSON file (Default=None).
    :param out_csv_file: if specified the generated metrics and confusion matrix are written to
                         a CSV file (Default=None).

    """
    import rsgislib
    import rsgislib.vectorutils
    import rsgislib.rastergis
    import rsgislib.rastergis.ratutils

    # Read columns from vector file.
    ref_vals = numpy.array(rsgislib.vectorutils.readVecColumn(in_vec_file, in_vec_lyr, ref_col))
    cls_vals = numpy.array(rsgislib.vectorutils.readVecColumn(in_vec_file, in_vec_lyr, cls_col))

    # Find unique class values
    unq_cls_names = numpy.unique(numpy.concatenate((numpy.unique(ref_vals), numpy.unique(cls_vals))))

    # Create LUTs assigning each class a unique int ID.
    cls_name_lut = dict()
    cls_id_lut = dict()
    for cls_id, cls_name in enumerate(unq_cls_names):
        cls_name_lut[cls_name] = cls_id
        cls_id_lut[cls_id] = cls_name

    # Create cls_id arrays
    ref_int_vals = numpy.zeros_like(ref_vals, dtype=int)
    cls_int_vals = numpy.zeros_like(cls_vals, dtype=int)
    for cls_name in unq_cls_names:
        ref_int_vals[ref_vals == cls_name] = cls_name_lut[cls_name]
        cls_int_vals[cls_vals == cls_name] = cls_name_lut[cls_name]

    try:
        rat_cols = rsgislib.rastergis.getRATColumns(cls_img)
    except:
        raise Exception("The input image does not have a RAT...")

    if img_cls_name_col not in rat_cols:
        raise Exception("The RAT does not contain the class name column specified ('{}')".format(img_cls_name_col))
    if img_hist_col not in rat_cols:
        raise Exception("The RAT does not contain the histogram column specified ('{}')".format(img_hist_col))

    img_hist_data = rsgislib.rastergis.ratutils.getColumnData(cls_img, img_hist_col)
    img_clsname_data = rsgislib.rastergis.ratutils.getColumnData(cls_img, img_cls_name_col)
    img_clsname_data[0] = ''

    rsgis_utils = rsgislib.RSGISPyUtils()
    pxl_size_x, pxl_size_y = rsgis_utils.getImageRes(cls_img)
    pxl_area = pxl_size_x * pxl_size_y

    # Find the class areas (pixel counts)
    cls_pxl_count_dict = dict()
    cls_area_dict = dict()
    tot_area = 0.0
    cls_pxl_counts = numpy.zeros_like(unq_cls_names, dtype=int)
    for i, cls_name in enumerate(img_clsname_data):
        cls_name = str(cls_name.decode())
        if (i > 0) and (cls_name !=''):
            if cls_name not in unq_cls_names:
                raise Exception("Class ('{}') found in image which was not in point samples...".format(cls_name))
            cls_pxl_counts[cls_name_lut[cls_name]] = img_hist_data[i]
            cls_pxl_count_dict[cls_name] = img_hist_data[i]
            cls_area_dict[cls_name] = img_hist_data[i] * pxl_area
            tot_area = tot_area + (img_hist_data[i] * pxl_area)

    acc_metrics = calc_class_accuracy_metrics(ref_int_vals, cls_int_vals, cls_pxl_counts, unq_cls_names)

    acc_metrics['pixel_count'] = cls_pxl_count_dict
    acc_metrics['pixel_area'] = cls_area_dict

    if out_json_file is not None:
        import json
        with open(out_json_file, 'w') as out_json_file_obj:
            json.dump(acc_metrics, out_json_file_obj, sort_keys=True, indent=4, separators=(',', ': '), ensure_ascii=False)

    if out_csv_file is not None:
        import csv
        with open(out_csv_file, mode='w') as out_csv_file_obj:
            acc_metrics_writer = csv.writer(out_csv_file_obj, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)

            # Overall Accuracy
            acc_metrics_writer.writerow(['overall accuracy', acc_metrics['accuracy']])
            acc_metrics_writer.writerow(['cohen kappa', acc_metrics['cohen_kappa']])
            acc_metrics_writer.writerow([''])

            # Quantity Metrics
            acc_metrics_writer.writerow(
                ['Allocation Disagreement (A)', acc_metrics['quantity_metrics']['Allocation Disagreement (A)']])
            acc_metrics_writer.writerow(
                ['Quantity Disagreement (Q)', acc_metrics['quantity_metrics']['Quantity Disagreement (Q)']])
            acc_metrics_writer.writerow(
                ['Proportion Correct (C)', acc_metrics['quantity_metrics']['Proportion Correct (C)']])
            acc_metrics_writer.writerow(
                ['Total Disagreement (D)', acc_metrics['quantity_metrics']['Total Disagreement (D)']])
            acc_metrics_writer.writerow([''])

            # Individual Class Scores
            acc_metrics_writer.writerow(['class', 'f1-score', 'precision', 'recall', 'support'])
            for cls_name in unq_cls_names:
                acc_metrics_writer.writerow(
                    [cls_name, acc_metrics[cls_name]['f1-score'], acc_metrics[cls_name]['precision'],
                     acc_metrics[cls_name]['recall'], acc_metrics[cls_name]['support']])
            # Overall macro and weighted
            acc_metrics_writer.writerow([''])
            acc_metrics_writer.writerow(
                ['macro avg', acc_metrics['macro avg']['f1-score'], acc_metrics['macro avg']['precision'],
                 acc_metrics['macro avg']['recall'], acc_metrics['macro avg']['support']])
            acc_metrics_writer.writerow(
                ['weighted avg', acc_metrics['weighted avg']['f1-score'], acc_metrics['weighted avg']['precision'],
                 acc_metrics['weighted avg']['recall'], acc_metrics['weighted avg']['support']])
            acc_metrics_writer.writerow([''])

            # Output the confusion matrix
            acc_metrics_writer.writerow(['Point Count Confusion Matrix'])
            cm_top_row = ['']
            for cls_name in unq_cls_names:
                cm_top_row.append(cls_name)
            cm_top_row.append('User Acc')
            acc_metrics_writer.writerow(cm_top_row)
            for cls_name, cm_row, user_acc in zip(unq_cls_names, acc_metrics['confusion_matrix'],
                                                  acc_metrics['user_accuracy']):
                row = [cls_name]
                for val in cm_row:
                    row.append(val)
                row.append(user_acc)
                acc_metrics_writer.writerow(row)
            cm_bot_row = ['Producer']
            for prod_val in acc_metrics['producer_accuracy']:
                cm_bot_row.append(prod_val)
            acc_metrics_writer.writerow(cm_bot_row)
            acc_metrics_writer.writerow([''])

            acc_metrics_writer.writerow(['Normalised Confusion Matrix'])
            # Output the normalised confusion matrix
            cm_top_row = ['']
            for cls_name in unq_cls_names:
                cm_top_row.append(cls_name)
            acc_metrics_writer.writerow(cm_top_row)
            for cls_name, cm_row in zip(unq_cls_names, acc_metrics['norm_confusion_matrix']):
                row = [cls_name]
                for val in cm_row:
                    row.append(val)
                acc_metrics_writer.writerow(row)

            acc_metrics_writer.writerow([''])
            acc_metrics_writer.writerow(['class', 'commission', 'ommision'])
            for i, cls_name in enumerate(unq_cls_names):
                acc_metrics_writer.writerow([cls_name, acc_metrics['commission'][i], acc_metrics['ommission'][i]])

            acc_metrics_writer.writerow([''])
            acc_metrics_writer.writerow(['class', 'pixel count', 'pixel area', 'Est. Prop. Area', 'Est. Area'])
            for i, cls_name in enumerate(unq_cls_names):
                acc_metrics_writer.writerow([cls_name, cls_pxl_count_dict[cls_name], cls_area_dict[cls_name],
                                             acc_metrics['est_prop_cls_area'][i],
                                             (tot_area * acc_metrics['est_prop_cls_area'][i])])

    if (out_json_file is None) and (out_csv_file is None):
        import pprint
        pprint.pprint(acc_metrics)



#!/usr/bin/env python
"""
The tools.filemanipulation module contains functions for manipulating and moving files around.
"""

# Import modules
import rsgislib

import glob
import os.path
import os
import shutil

def sortImgsUTM2DIRs(inputImgsDIR, fileSearchStr, outBaseDIR):
    """
A function which will sort a series of input image files which
a projected using the UTM system into individual directories per
UTM zone. Please note that the input files are moved on your system!!

Where:

:param inputImgsDIR: directory where the input files are to be found.
:param fileSearchStr: the wildcard search string to find files within the input directory (e.g., \*.kea).
:param outBaseDIR: the output directory where the UTM folders will be created and the files copied.

"""
    rsgisUtils = rsgislib.RSGISPyUtils()
    inFiles = glob.glob(os.path.join(inputImgsDIR, fileSearchStr))
    for imgFile in inFiles:
        utmZone = rsgisUtils.getUTMZone(imgFile)
        if utmZone is not None:
            outDIR = os.path.join(outBaseDIR, 'utm'+utmZone)
            if not os.path.exists(outDIR):
                os.makedirs(outDIR)
            imgFileList = rsgisUtils.getImageFiles(imgFile)
            for tmpFile in imgFileList:
                print('Moving: ' + tmpFile)
                outFile = os.path.join(outDIR, os.path.basename(tmpFile))
                shutil.move(tmpFile, outFile)


def createSHA1Hash(input_file, block_size=4096):
    """
    A function which calculates finds the SHA1 hash string of the input file.
    :param input_file: the input file for which the SHA1 hash string with be found.
    :param block_size: the size of the blocks the file is read in in bytes (default 4096; i.e., 4kb)
    :return: SHA1 hash string of the file.

    """
    import hashlib
    sha1_hash = hashlib.sha1()
    with open(input_file, "rb") as f:
        # Read and update hash string value in blocks
        for byte_block in iter(lambda: f.read(block_size), b""):
            sha1_hash.update(byte_block)
    return sha1_hash.hexdigest()


def createSHA224Hash(input_file, block_size=4096):
    """
    A function which calculates finds the SHA224 hash string of the input file.
    :param input_file: the input file for which the SHA224 hash string with be found.
    :param block_size: the size of the blocks the file is read in in bytes (default 4096; i.e., 4kb)
    :return: SHA224 hash string of the file.

    """
    import hashlib
    sha224_hash = hashlib.sha224()
    with open(input_file, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(block_size), b""):
            sha224_hash.update(byte_block)
    return sha224_hash.hexdigest()


def createSHA256Hash(input_file, block_size=4096):
    """
    A function which calculates finds the SHA256 hash string of the input file.
    :param input_file: the input file for which the SHA256 hash string with be found.
    :param block_size: the size of the blocks the file is read in in bytes (default 4096; i.e., 4kb)
    :return: SHA256 hash string of the file.

    """
    import hashlib
    sha256_hash = hashlib.sha256()
    with open(input_file, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(block_size), b""):
            sha256_hash.update(byte_block)
    return sha256_hash.hexdigest()


def createSHA384Hash(input_file, block_size=4096):
    """
    A function which calculates finds the SHA384 hash string of the input file.
    :param input_file: the input file for which the SHA384 hash string with be found.
    :param block_size: the size of the blocks the file is read in in bytes (default 4096; i.e., 4kb)
    :return: SHA384 hash string of the file.

    """
    import hashlib
    sha384_hash = hashlib.sha384()
    with open(input_file, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(block_size), b""):
            sha384_hash.update(byte_block)
    return sha384_hash.hexdigest()


def createSHA512Hash(input_file, block_size=4096):
    """
    A function which calculates finds the SHA512 hash string of the input file.
    :param input_file: the input file for which the SHA512 hash string with be found.
    :param block_size: the size of the blocks the file is read in in bytes (default 4096; i.e., 4kb)
    :return: SHA512 hash string of the file.

    """
    import hashlib
    sha512_hash = hashlib.sha512()
    with open(input_file, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(block_size), b""):
            sha512_hash.update(byte_block)
    return sha512_hash.hexdigest()


def createMD5Hash(input_file, block_size=4096):
    """
    A function which calculates finds the MD5 hash string of the input file.
    :param input_file: the input file for which the MD5 hash string with be found.
    :param block_size: the size of the blocks the file is read in in bytes (default 4096; i.e., 4kb)
    :return: MD5 hash string of the file.

    """
    import hashlib
    md5_hash = hashlib.md5()
    with open(input_file, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(block_size), b""):
            md5_hash.update(byte_block)
    return md5_hash.hexdigest()


def createBlake2BHash(input_file, block_size=4096):
    """
    A function which calculates finds the Blake2B hash string of the input file.
    :param input_file: the input file for which the Blake2B hash string with be found.
    :param block_size: the size of the blocks the file is read in in bytes (default 4096; i.e., 4kb)
    :return: Blake2B hash string of the file.

    """
    import hashlib
    blake2b_hash = hashlib.blake2b()
    with open(input_file, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(block_size), b""):
            blake2b_hash.update(byte_block)
    return blake2b_hash.hexdigest()


def createBlake2SHash(input_file, block_size=4096):
    """
    A function which calculates finds the Blake2S hash string of the input file.
    :param input_file: the input file for which the Blake2S hash string with be found.
    :param block_size: the size of the blocks the file is read in in bytes (default 4096; i.e., 4kb)
    :return: Blake2S hash string of the file.

    """
    import hashlib
    blake2s_hash = hashlib.blake2s()
    with open(input_file, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(block_size), b""):
            blake2s_hash.update(byte_block)
    return blake2s_hash.hexdigest()


def createSHA3_224Hash(input_file, block_size=4096):
    """
    A function which calculates finds the SHA3_224 hash string of the input file.
    :param input_file: the input file for which the SHA3_224 hash string with be found.
    :param block_size: the size of the blocks the file is read in in bytes (default 4096; i.e., 4kb)
    :return: SHA3_224 hash string of the file.

    """
    import hashlib
    sha224_hash = hashlib.sha3_224()
    with open(input_file, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(block_size), b""):
            sha224_hash.update(byte_block)
    return sha224_hash.hexdigest()


def createSHA3_256Hash(input_file, block_size=4096):
    """
    A function which calculates finds the SHA3_256 hash string of the input file.
    :param input_file: the input file for which the SHA3_256 hash string with be found.
    :param block_size: the size of the blocks the file is read in in bytes (default 4096; i.e., 4kb)
    :return: SHA3_256 hash string of the file.

    """
    import hashlib
    sha256_hash = hashlib.sha3_256()
    with open(input_file, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(block_size), b""):
            sha256_hash.update(byte_block)
    return sha256_hash.hexdigest()


def createSHA3_384Hash(input_file, block_size=4096):
    """
    A function which calculates finds the SHA3_384 hash string of the input file.
    :param input_file: the input file for which the SHA3_384 hash string with be found.
    :param block_size: the size of the blocks the file is read in in bytes (default 4096; i.e., 4kb)
    :return: SHA3_384 hash string of the file.

    """
    import hashlib
    sha384_hash = hashlib.sha3_384()
    with open(input_file, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(block_size), b""):
            sha384_hash.update(byte_block)
    return sha384_hash.hexdigest()


def createSHA3_512Hash(input_file, block_size=4096):
    """
    A function which calculates finds the SHA3_512 hash string of the input file.
    :param input_file: the input file for which the SHA3_512 hash string with be found.
    :param block_size: the size of the blocks the file is read in in bytes (default 4096; i.e., 4kb)
    :return: SHA3_512 hash string of the file.

    """
    import hashlib
    sha512_hash = hashlib.sha3_512()
    with open(input_file, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(block_size), b""):
            sha512_hash.update(byte_block)
    return sha512_hash.hexdigest()





#!/usr/bin/env python
############################################################################
#  specunmixing.py
#
#  Copyright 2020 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Perform spectral unmixing
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 07/01/2017
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
############################################################################

# UCLS, NNLS, FCLS classes adapted from pysptools module:
# https://github.com/ctherien/pysptools
# which has has the following license:
# ------------------------------------------------------------------------------
# Copyright (c) 2013-2014, Christian Therien
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ------------------------------------------------------------------------------
#
from abc import ABCMeta, abstractmethod
import numpy
from rios import applier
from rios import cuiprogress
import rsgislib


class UnmixingImpl(object):
    """
    An abstract class which can be used to implement various
    spectral unmixing implementations and use through the same
    interface.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def calc_abundance(self, M, U):
        """
        An abstract function which implements a method for spectral unmixing.
        N is the number of image pixels being processed. 
        p is the number of image bands
        q is the number of endmembers.

        :param M: `numpy array` 2D data matrix (N x p).
        :param U: `numpy array` 2D matrix of endmembers (q x p).
        :return: `numpy array` An abundance maps (N x q).
        """
        pass

    def read_endmembers_mtxt(self, endmembers_file):
        """
        A function to read endmembers mtxt file
        :param endmembers_file: File path to the mtxt input file.
        :return: [m, n, endmemarr]; m size, n size and a numpy array with the end members
        """
        dataList = []
        dataFile = open(endmembers_file, 'r')
        for line in dataFile:
            line = line.strip()
            if line != "":
                dataList.append(line)
        dataFile.close()
        m = int(dataList[0].split('=')[-1])
        n = int(dataList[1].split('=')[-1])
        endmemarr = numpy.zeros((m, n))
        datavals = dataList[2].split(',')

        i = 0
        for m_idx in range(m):
            for n_idx in range(n):
                endmemarr[m_idx, n_idx] = float(datavals[i])
                i = i + 1

        return m, n, endmemarr

    def read_endmembers_mtxt_weight(self, endmembers_file, weight=100):
        """
        A function to read endmembers mtxt file
        :param endmembers_file: File path to the mtxt input file.
        :return: [m, n, endmemarr]; m size, n size and a numpy array with the end members
        """
        dataList = []
        dataFile = open(endmembers_file, 'r')
        for line in dataFile:
            line = line.strip()
            if line != "":
                dataList.append(line)
        dataFile.close()
        m = int(dataList[0].split('=')[-1])
        n = int(dataList[1].split('=')[-1]) + 1
        endmemarr = numpy.zeros((m, n))
        datavals = dataList[2].split(',')

        i = 0
        for m_idx in range(m):
            for n_idx in range(n - 1):
                endmemarr[m_idx, n_idx] = float(datavals[i])
                i = i + 1
            endmemarr[m_idx, n - 1] = weight

        return m, n, endmemarr

    def perform_simple_unmixing(self, input_image, output_image, gdalformat, endmembers_file, weight=None,
                                scale_factor=1000):
        """
        A function which uses the RIOS to iterate through the input image
        and perform a simple/standard spectral unmixing on the input image
        using the calc_abundance function.

        :param input_image: The file path to a GDAL readable input image.
        :param output_image: The file path to the GDAL writable output image
                             (Note pixel values will be between 0-1000)
        :param gdalformat: The output image format to be used.
        :param endmembers_file: The endmembers for the unmixing in the RSGISLib mtxt format.
        :param weight: Optional (if None ignored) to provide a weight to implement the approach of Scarth et al (2010)
                   adding a weight to the least squares optimisation to get the abundances to sum to 1.
        :param scale_factor: Scale factor for integerising the resulting image. If value is 1 then output image
                             will be a floating point image.

        References:
            Scarth, P., Roder, A., & Schmidt, M. (2010). Tracking grazing pressure and climate
            interaction—The role of Landsat fractional cover in time series analysis. Proceedings of
            the 15th Australasian Remote Sensing and Photogrammetry ConferenceAustralia: Alice Springs.
            http://dx.doi.org/10.6084/m9.ﬁgshare.94250.

        """
        try:
            import tqdm
            progress_bar = rsgislib.TQDMProgressBar()
        except:
            progress_bar = cuiprogress.GDALProgressBar()

        if weight is not None:
            endmembers_info = self.read_endmembers_mtxt_weight(endmembers_file, weight)
        else:
            endmembers_info = self.read_endmembers_mtxt(endmembers_file)
        print(endmembers_info)

        infiles = applier.FilenameAssociations()
        infiles.image = input_image
        outfiles = applier.FilenameAssociations()
        outfiles.outimage = output_image
        otherargs = applier.OtherInputs()
        otherargs.endmembers_q = endmembers_info[0]
        otherargs.endmembers_p = endmembers_info[1]
        otherargs.endmembers_arr = endmembers_info[2]
        otherargs.weight = weight
        otherargs.scale_factor = scale_factor
        aControls = applier.ApplierControls()
        aControls.progress = progress_bar
        aControls.drivername = gdalformat
        aControls.omitPyramids = True
        aControls.calcStats = False

        def _simple_unmix(info, inputs, outputs, otherargs):
            """
            This is an internal rios function
            """
            block_shp = inputs.image.shape
            img_flat = inputs.image.reshape([block_shp[0], (block_shp[1] * block_shp[2])]).T
            if otherargs.weight is not None:
                img_flat_dtype = img_flat.dtype
                tmp_img_flat = img_flat
                img_flat = numpy.zeros((img_flat.shape[0], img_flat.shape[1] + 1), dtype=img_flat_dtype)
                img_flat[...] = weight
                img_flat[:, :-1] = tmp_img_flat
            img_flat_shp = img_flat.shape

            img_nodata = numpy.where(img_flat == 0, True, False)
            img_flat_nodata = numpy.all(img_nodata, axis=1)

            ID = numpy.arange(img_flat_shp[0])
            ID = ID[img_flat_nodata == False]
            img_flat_data = img_flat[img_flat_nodata == False, ...]

            abundances_arr = self.calc_abundance(img_flat_data, otherargs.endmembers_arr)
            if otherargs.scale_factor > 1:
                outarr = numpy.zeros([img_flat_shp[0], otherargs.endmembers_q], dtype=numpy.int16)
            else:
                outarr = numpy.zeros([img_flat_shp[0], otherargs.endmembers_q], dtype=numpy.float32)
            outarr[ID] = (abundances_arr * otherargs.scale_factor)
            outarr = outarr.T
            outputs.outimage = outarr.reshape((otherargs.endmembers_q, block_shp[1], block_shp[2]))

        applier.apply(_simple_unmix, infiles, outfiles, otherargs, controls=aControls)

    def predict_refl_img_from_simple_unmixing(self, input_unmix_img, endmembers_file, output_img, gdalformat,
                                              scale_factor=1000):
        """
        """
        try:
            import tqdm
            progress_bar = rsgislib.TQDMProgressBar()
        except:
            progress_bar = cuiprogress.GDALProgressBar()

        endmembers_info = self.read_endmembers_mtxt(endmembers_file)

        infiles = applier.FilenameAssociations()
        infiles.image_unmix = input_unmix_img
        outfiles = applier.FilenameAssociations()
        outfiles.outimage = output_img
        otherargs = applier.OtherInputs()
        otherargs.endmembers_q = endmembers_info[0]
        otherargs.endmembers_p = endmembers_info[1]
        otherargs.endmembers_arr = endmembers_info[2]
        otherargs.scale_factor = scale_factor
        aControls = applier.ApplierControls()
        aControls.progress = progress_bar
        aControls.drivername = gdalformat
        aControls.omitPyramids = True
        aControls.calcStats = False

        def _predict_refl_img(info, inputs, outputs, otherargs):
            """
            This is an internal rios function
            """
            block_unmix_shp = inputs.image_unmix.shape
            img_unmix_coef_flat = inputs.image_unmix.reshape(
                [block_unmix_shp[0], (block_unmix_shp[1] * block_unmix_shp[2])]).T

            img_unmix_coef_flat_data_flt = numpy.zeros_like(img_unmix_coef_flat, dtype=numpy.float32)
            img_unmix_coef_flat_data_flt[...] = img_unmix_coef_flat / float(scale_factor)

            img_nodata_flat_shp = img_unmix_coef_flat_data_flt.shape

            pred_refl_img = numpy.zeros((img_unmix_coef_flat.shape[0], otherargs.endmembers_p), dtype=numpy.int16)
            for i in range(img_unmix_coef_flat.shape[0]):
                for q in range(otherargs.endmembers_q):
                    pred_refl_img[i] = pred_refl_img[i] + (
                                img_unmix_coef_flat_data_flt[i, q] * otherargs.endmembers_arr[q])

            pred_refl_img = pred_refl_img.T
            outputs.outimage = pred_refl_img.reshape((otherargs.endmembers_p, block_unmix_shp[1], block_unmix_shp[2]))

        applier.apply(_predict_refl_img, infiles, outfiles, otherargs, controls=aControls)

    def calc_unmixing_rmse_residualerr(self, input_img, input_unmix_img, endmembers_file, output_img, gdalformat,
                                       scale_factor=1000):
        """
        """
        try:
            import tqdm
            progress_bar = rsgislib.TQDMProgressBar()
        except:
            progress_bar = cuiprogress.GDALProgressBar()

        endmembers_info = self.read_endmembers_mtxt(endmembers_file)

        infiles = applier.FilenameAssociations()
        infiles.image_orig = input_img
        infiles.image_unmix = input_unmix_img
        outfiles = applier.FilenameAssociations()
        outfiles.outimage = output_img
        otherargs = applier.OtherInputs()
        otherargs.endmembers_q = endmembers_info[0]
        otherargs.endmembers_p = endmembers_info[1]
        otherargs.endmembers_arr = endmembers_info[2]
        otherargs.scale_factor = scale_factor
        aControls = applier.ApplierControls()
        aControls.progress = progress_bar
        aControls.drivername = gdalformat
        aControls.omitPyramids = True
        aControls.calcStats = False

        def _calc_unmix_err_rmse(info, inputs, outputs, otherargs):
            """
            This is an internal rios function
            """
            block_refl_shp = inputs.image_orig.shape
            img_orig_refl_flat = inputs.image_orig.reshape(
                [block_refl_shp[0], (block_refl_shp[1] * block_refl_shp[2])]).T
            block_unmix_shp = inputs.image_unmix.shape
            img_unmix_coef_flat = inputs.image_unmix.reshape(
                [block_unmix_shp[0], (block_unmix_shp[1] * block_unmix_shp[2])]).T
            img_flat_shp = img_orig_refl_flat.shape

            img_orig_refl_nodata = numpy.where(img_orig_refl_flat == 0, True, False)
            img_orig_refl_flat_nodata = numpy.all(img_orig_refl_nodata, axis=1)

            ID = numpy.arange(img_flat_shp[0])
            ID = ID[img_orig_refl_flat_nodata == False]
            img_orig_refl_flat_data = img_orig_refl_flat[img_orig_refl_flat_nodata == False, ...]
            img_orig_refl_flat_data_flt = numpy.zeros_like(img_orig_refl_flat_data, dtype=numpy.float32)
            img_orig_refl_flat_data_flt[...] = img_orig_refl_flat_data
            img_unmix_coef_flat_data = img_unmix_coef_flat[img_orig_refl_flat_nodata == False, ...]
            img_unmix_coef_flat_data_flt = numpy.zeros_like(img_unmix_coef_flat_data, dtype=numpy.float32)
            img_unmix_coef_flat_data_flt[...] = img_unmix_coef_flat_data / float(scale_factor)

            img_nodata_flat_shp = img_unmix_coef_flat_data_flt.shape

            pred_refl_img = numpy.zeros_like(img_orig_refl_flat_data, dtype=float)
            for i in range(img_nodata_flat_shp[0]):
                for q in range(otherargs.endmembers_q):
                    pred_refl_img[i] = pred_refl_img[i] + (
                                img_unmix_coef_flat_data_flt[i, q] * otherargs.endmembers_arr[q])

            # Calc Diff
            diff = img_orig_refl_flat_data_flt - pred_refl_img

            # Calc RMSE
            diff_sq = diff * diff
            mean_sum_diff_sq = numpy.sum(diff_sq, axis=1) / otherargs.endmembers_p
            rmse_arr = numpy.sqrt(mean_sum_diff_sq)

            # Calc Residual Error
            abs_diff = numpy.absolute(diff)
            residual_arr = numpy.sum(abs_diff, axis=1) / otherargs.endmembers_p

            outarr = numpy.zeros((img_flat_shp[0], 2))
            outarr[ID] = numpy.stack([rmse_arr, residual_arr], axis=-1)
            outarr = outarr.T
            outputs.outimage = outarr.reshape((2, block_refl_shp[1], block_refl_shp[2]))

        applier.apply(_calc_unmix_err_rmse, infiles, outfiles, otherargs, controls=aControls)


class Unmixing_UCLS(UnmixingImpl):

    def calc_abundance(self, M, U):
        """
        Performs unconstrained least squares abundance estimation.

        :param M: `numpy array` 2D data matrix (N x p).
        :param U: `numpy array` 2D matrix of endmembers (q x p).
        :return: `numpy array` An abundance maps (N x q).
        """
        Uinv = numpy.linalg.pinv(U.T)
        return numpy.dot(Uinv, M[0:, :].T).T


class Unmixing_NNLS(UnmixingImpl):

    def calc_abundance(self, M, U):
        """
        NNLS performs non-negative constrained least squares of each pixel
        in M using the endmember signatures of U.  Non-negative constrained least
        squares with the abundance nonnegative constraint (ANC).
        Utilizes the method of Bro and de Jong (2007).

        References:
            Bro R., de Jong S., Journal of Chemometrics, 1997, 11, 393-401.

        :param M: `numpy array` 2D data matrix (N x p).
        :param U: `numpy array` 2D matrix of endmembers (q x p).
        :return: `numpy array` An abundance maps (N x q).

        """
        import scipy.optimize as opt

        N, p1 = M.shape
        q, p2 = U.shape

        X = numpy.zeros((N, q), dtype=numpy.float32)
        MtM = numpy.dot(U, U.T)
        for n1 in range(N):
            # opt.nnls() return a tuple, the first element is the result
            X[n1] = opt.nnls(MtM, numpy.dot(U, M[n1]))[0]
        return X


class Unmixing_FCLS(UnmixingImpl):

    def _numpy_None_vstack(self, A1, A2):
        if A1 is None:
            return A2
        else:
            return numpy.vstack([A1, A2])

    def _numpy_None_concatenate(self, A1, A2):
        if A1 is None:
            return A2
        else:
            return numpy.concatenate([A1, A2])

    def _numpy_to_cvxopt_matrix(self, A):
        from cvxopt import matrix
        A = numpy.array(A, dtype=numpy.float64)
        if A.ndim == 1:
            return matrix(A, (A.shape[0], 1), 'd')
        else:
            return matrix(A, A.shape, 'd')

    def calc_abundance(self, M, U):
        """
        Performs fully constrained least squares of each pixel in M
        using the endmember signatures of U. Fully constrained least squares
        is least squares with the abundance sum-to-one constraint (ASC) and the
        abundance nonnegative constraint (ANC).

        References:
             Daniel Heinz, Chein-I Chang, and Mark L.G. Fully Constrained
             Least-Squares Based Linear Unmixing. Althouse. IEEE. 1999.

        Notes:
             This code was taken from https://github.com/ctherien/pysptools, which
             used the following threes sources to build this implementation of the
             algorithm:
                 * The function hyperFclsMatlab, part of the Matlab Hyperspectral
                   Toolbox of Isaac Gerg.
                 * The Matlab (tm) help on lsqlin.
                 * And the Python implementation of lsqlin by Valera Vishnevskiy:
                   http://maggotroot.blogspot.ca/2013/11/constrained-linear-least-squares-in.html

        :param M: `numpy array` 2D data matrix (N x p).
        :param U: `numpy array` 2D matrix of endmembers (q x p).
        :return: `numpy array` An abundance maps (N x q).

        """
        from cvxopt import solvers, matrix
        solvers.options['show_progress'] = False
        N, p1 = M.shape
        nvars, p2 = U.shape

        C = self._numpy_to_cvxopt_matrix(U.T)
        Q = C.T * C

        lb_A = -numpy.eye(nvars)
        lb = numpy.repeat(0, nvars)
        A = self._numpy_None_vstack(None, lb_A)
        b = self._numpy_None_concatenate(None, -lb)
        A = self._numpy_to_cvxopt_matrix(A)
        b = self._numpy_to_cvxopt_matrix(b)

        Aeq = self._numpy_to_cvxopt_matrix(numpy.ones((1, nvars)))
        beq = self._numpy_to_cvxopt_matrix(numpy.ones(1))

        M = numpy.array(M, dtype=numpy.float64)
        X = numpy.zeros((N, nvars), dtype=numpy.float32)
        for n1 in range(N):
            d = matrix(M[n1], (p1, 1), 'd')
            q = - d.T * C
            sol = solvers.qp(Q, q.T, A, b, Aeq, beq, None, None)['x']
            X[n1] = numpy.array(sol).squeeze()
        return X


def rescale_unmixing_results(input_img, output_img, gdalformat, scale_factor=1000):
    """
    A function which rescales an output from a spectral unmixing
    (e.g., rsgislib.imagecalc.specunmixing.calc_unconstrained_unmixing) so that
    negative values are removed and each pixel sums to 1.

    :param inputImg: Input image with the spectral unmixing result (pixels need to range from 0-1)
    :param outputImg: Output image with the result of the rescaling (pixel values will be in range 0-1)
    :param gdalformat: the file format of the output file.

    """
    try:
        import tqdm
        progress_bar = rsgislib.TQDMProgressBar()
    except:
        progress_bar = cuiprogress.GDALProgressBar()

    infiles = applier.FilenameAssociations()
    infiles.image = input_img
    outfiles = applier.FilenameAssociations()
    outfiles.outimage = output_img
    otherargs = applier.OtherInputs()
    otherargs.scale_factor = scale_factor
    aControls = applier.ApplierControls()
    aControls.progress = progress_bar
    aControls.drivername = gdalformat
    aControls.omitPyramids = True
    aControls.calcStats = False

    def _applyUnmixRescale(info, inputs, outputs, otherargs):
        """
        This is an internal rios function
        """
        inputs.image[inputs.image < 0] = 0
        if otherargs.scale_factor == 1:
            outputs.outimage = numpy.zeros_like(inputs.image, dtype=numpy.float32)
        else:
            outputs.outimage = numpy.zeros_like(inputs.image, dtype=numpy.uint16)
        for idx in range(inputs.image.shape[0]):
            outputs.outimage[idx] = (inputs.image[idx] / numpy.sum(inputs.image, axis=0)) * otherargs.scale_factor

    applier.apply(_applyUnmixRescale, infiles, outfiles, otherargs, controls=aControls)


def calc_unconstrained_unmixing(input_img, output_img, gdalformat, endmembers_file, weight=None,
                                rescaled_output_img=None, calc_stats=True, scale_factor=1000):
    """
    A function which uses the Unmixing_UCLS class to perform an uncontrained spectral
    unmixing.

    :param input_image: The file path to a GDAL readable input image.
    :param output_image: The file path to the GDAL writable output image
                         (Note pixel values will be between 0-1000)
    :param gdalformat: The output image format to be used.
    :param endmembers_file: The endmembers for the unmixing in the RSGISLib mtxt format.
    :param weight: Optional (if None ignored) to provide a weight to implement the approach of Scarth et al (2010)
                   adding a weight to the least squares optimisation to get the abundances to sum to 1.
    :param rescaled_output_img: A file path to an output image which has been rescaled so the abundances sum to 1.
                                Optional, if None then ignored.
    :param calc_stats: If True (default) then image statistics and pyramids are calculated for the output images
    :param scale_factor: Scale factor for integerising the resulting image. If value is 1 then output image
                             will be a floating point image.

    References:
        Scarth, P., Roder, A., & Schmidt, M. (2010). Tracking grazing pressure and climate
        interaction—The role of Landsat fractional cover in time series analysis. Proceedings of
        the 15th Australasian Remote Sensing and Photogrammetry ConferenceAustralia: Alice Springs.
        http://dx.doi.org/10.6084/m9.ﬁgshare.94250.

    """
    unmix_obj = Unmixing_UCLS()
    unmix_obj.perform_simple_unmixing(input_img, output_img, gdalformat, endmembers_file, weight, scale_factor)
    if calc_stats:
        import rsgislib.imageutils
        rsgislib.imageutils.popImageStats(output_img, usenodataval=True, nodataval=(scale_factor + 1),
                                          calcpyramids=True)

    if rescaled_output_img is not None:
        rescale_unmixing_results(output_img, rescaled_output_img, gdalformat, scale_factor=scale_factor)
        if calc_stats:
            rsgislib.imageutils.popImageStats(rescaled_output_img, usenodataval=True, nodataval=(scale_factor + 1),
                                              calcpyramids=True)


def calc_non_negative_unmixing(input_img, output_img, gdalformat, endmembers_file, weight=None,
                               rescaled_output_img=None, calc_stats=True, scale_factor=1000):
    """
    A function which uses the Unmixing_NNLS class to perform an non-negative constrained
    spectral unmixing.

    :param input_image: The file path to a GDAL readable input image.
    :param output_image: The file path to the GDAL writable output image
                         (Note pixel values will be between 0-1000)
    :param gdalformat: The output image format to be used.
    :param endmembers_file: The endmembers for the unmixing in the RSGISLib mtxt format.
    :param weight: Optional (if None ignored) to provide a weight to implement the approach of Scarth et al (2010)
                   adding a weight to the least squares optimisation to get the abundances to sum to 1.
    :param rescaled_output_img: A file path to an output image which has been rescaled so the abundances sum to 1.
                                Optional, if None then ignored.
    :param calc_stats: If True (default) then image statistics and pyramids are calculated for the output images
    :param scale_factor: Scale factor for integerising the resulting image. If value is 1 then output image
                             will be a floating point image.

    References:
        Scarth, P., Roder, A., & Schmidt, M. (2010). Tracking grazing pressure and climate
        interaction—The role of Landsat fractional cover in time series analysis. Proceedings of
        the 15th Australasian Remote Sensing and Photogrammetry ConferenceAustralia: Alice Springs.
        http://dx.doi.org/10.6084/m9.ﬁgshare.94250.

    """
    unmix_obj = Unmixing_NNLS()
    unmix_obj.perform_simple_unmixing(input_img, output_img, gdalformat, endmembers_file, weight, scale_factor)
    if calc_stats:
        import rsgislib.imageutils
        rsgislib.imageutils.popImageStats(output_img, usenodataval=True, nodataval=(scale_factor + 1),
                                          calcpyramids=True)
    if rescaled_output_img is not None:
        rescale_unmixing_results(output_img, rescaled_output_img, gdalformat, scale_factor=scale_factor)
        if calc_stats:
            rsgislib.imageutils.popImageStats(rescaled_output_img, usenodataval=True, nodataval=(scale_factor + 1),
                                              calcpyramids=True)


def calc_fully_constrained_unmixing(input_img, output_img, gdalformat, endmembers_file, calc_stats=True,
                                    scale_factor=1000):
    """
    A function which uses the Unmixing_FCLS class to perform a fully contrained spectral
    unmixing.

    :param input_image: The file path to a GDAL readable input image.
    :param output_image: The file path to the GDAL writable output image
                         (Note pixel values will be between 0-1000)
    :param gdalformat: The output image format to be used.
    :param endmembers_file: The endmembers for the unmixing in the RSGISLib mtxt format.
    :param calc_stats: If True (default) then image statistics and pyramids are calculated for the output images
    :param scale_factor: Scale factor for integerising the resulting image. If value is 1 then output image
                             will be a floating point image.

    """
    unmix_obj = Unmixing_FCLS()
    unmix_obj.perform_simple_unmixing(input_img, output_img, gdalformat, endmembers_file, scale_factor=scale_factor)
    if calc_stats:
        import rsgislib.imageutils
        rsgislib.imageutils.popImageStats(output_img, usenodataval=True, nodataval=(scale_factor + 1),
                                          calcpyramids=True)


def calc_unmix_rmse_residualerr(input_refl_img, input_unmix_coef_img, endmembers_file, output_img, gdalformat,
                                scale_factor=1000, calc_stats=True):
    """
    A function to calculate the prediction residual and RMSE using the defined abundances and endmembers.

    :param input_refl_img: The original input reference image.
    :param input_unmix_coef_img: The unmixed abundance coefficients.
    :param endmembers_file: The endmembers file used for the unmixing in the RSGISLib mtxt format.
    :param output_image: The file path to the GDAL writable output image (Band 1: RMSE, Band 2: Residual)
    :param gdalformat: The output image format to be used.
    :param scale_factor: Scale factor used to integerising the endmember abundances image.
    :param calc_stats: If True (default) then image statistics and pyramids are calculated for the output images

    """
    unmix_obj = Unmixing_UCLS()
    unmix_obj.calc_unmixing_rmse_residualerr(input_refl_img, input_unmix_coef_img, endmembers_file, output_img,
                                             gdalformat, scale_factor)
    if calc_stats:
        import rsgislib.imageutils
        rsgislib.imageutils.setBandNames(output_img, ['RMSE', 'Residual'], feedback=False)
        rsgislib.imageutils.popImageStats(output_img, usenodataval=False, nodataval=0, calcpyramids=True)


def predict_refl_img_from_simple_unmixing(input_unmix_coef_img, endmembers_file, output_img, gdalformat,
                                          scale_factor=1000, calc_stats=True):
    """
    A function to calculate the predicted pixel value using the inputted abundances and endmembers.

    :param input_unmix_coef_img: The unmixed abundance coefficients.
    :param endmembers_file: The endmembers file used for the unmixing in the RSGISLib mtxt format.
    :param output_image: The file path to the GDAL writable output image
    :param gdalformat: The output image format to be used.
    :param scale_factor: Scale factor used to integerising the endmember abundances image.
    :param calc_stats: If True (default) then image statistics and pyramids are calculated for the output images

    """
    unmix_obj = Unmixing_UCLS()
    unmix_obj.predict_refl_img_from_simple_unmixing(input_unmix_coef_img, endmembers_file, output_img, gdalformat,
                                                    scale_factor)
    if calc_stats:
        import rsgislib.imageutils
        rsgislib.imageutils.popImageStats(output_img, usenodataval=True, nodataval=0, calcpyramids=True)





#!/usr/bin/env python
"""
The Species Distrubution Modelling (SDM) module provides a set of tools and utilities for
undertaking a species distribution modelling project.

These functions have been developed by Osian Roberts.

"""

from .pysdm import *




#! /usr/bin/env python
############################################################################
#  clustersklearn.py
#
#  Copyright 2020 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide a set of utilities which combine commands to create
#           useful extra functionality and make it more easily available
#           to be reused.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 09/01/2020
# Version: 1.0
#
# History:
# Version 1.0 - Refactored from previous structure.
#
###########################################################################

import rsgislib
import rsgislib.imageutils
import rsgislib.rastergis

import tqdm

import numpy

import osgeo.gdal as gdal

from sklearn.cluster import MiniBatchKMeans

from rios.imagereader import ImageReader
from rios.imagewriter import ImageWriter

def img_pixel_sample_cluster(inputImg, outputImg, gdalformat='KEA', noDataVal=0, imgSamp=100,
                             clusterer=MiniBatchKMeans(n_clusters=60, init='k-means++', max_iter=100, batch_size=100),
                             calcStats=True, useMeanShiftEstBandWidth=False):
    """
A function which allows a clustering to be performed using the algorithms available
within the scikit-learn library. The clusterer is trained on a sample of the input
image and then applied using the predict function (therefore this function is only
compatiable with clusterers which have the predict function implemented) to the whole
image.

:param inputImg: input image file.
:param outputImg: output image file.
:param gdalformat: output image file format.
:param noDataVal: no data value associated with the input image.
:param imgSamp: the input image sampling. (e.g., 100 is every 100th pixel)
:param clusterer: clusterer from scikit-learn which must have a predict function.
:param calcStats: calculate image pixel statistics, histogram and image pyramids - note if you are not using a
                  KEA file then the format needs to support RATs for this option as histogram and colour table
                  are written to RAT.
:param useMeanShiftEstBandWidth: use the mean-shift algorithm as the clusterer (pass None as the clusterer) where
                                 the bandwidth is calculated from the data itself.

"""
    print('Sample input image:')
    dataSamp = rsgislib.imageutils.extractImgPxlSample(inputImg, imgSamp, noDataVal)

    if useMeanShiftEstBandWidth:
        print('Using Mean-Shift predict bandwidth')
        from sklearn.cluster import MeanShift, estimate_bandwidth
        bandwidth = estimate_bandwidth(dataSamp, quantile=0.2, n_samples=500)
        clusterer = MeanShift(bandwidth=bandwidth, bin_seeding=True)

    print('Fit Clusterer')
    outClust = clusterer.fit(dataSamp)
    print('Fitted Clusterer')

    print('Apply to whole image:')
    reader = ImageReader(inputImg, windowxsize=200, windowysize=200)
    writer = None
    for (info, block) in tqdm.tqdm(reader):
        blkShape = block.shape
        blkBands = block.reshape((blkShape[0], (blkShape[1 ] *blkShape[2]))).T
        ID = numpy.arange(blkBands.shape[0])
        outClusterVals = numpy.zeros((blkBands.shape[0]))

        finiteMskArr = numpy.isfinite(blkBands).all(axis=1)
        ID = ID[finiteMskArr]
        blkBands = blkBands[finiteMskArr]

        noDataValArr = numpy.logical_not(numpy.where(blkBands == noDataVal, True, False).all(axis=1))

        blkBandsNoData = blkBands[noDataValArr]
        ID = ID[noDataValArr]

        if ID.shape[0] > 0:
            outPred = clusterer.predict(blkBandsNoData ) +1
            outClusterVals[ID] = outPred

        outClusterValsOutArr = outClusterVals.reshape([1 ,blkShape[1] ,blkShape[2]])

        if writer is None:
            writer = ImageWriter(outputImg, info=info, firstblock=outClusterValsOutArr, drivername=gdalformat, creationoptions=[])
        else:
            writer.write(outClusterValsOutArr)
    writer.close(calcStats=False)

    if calcStats:
        rsgislib.rastergis.populateStats(clumps=outputImg, addclrtab=True, calcpyramids=True, ignorezero=True)


def img_pixel_tiled_cluster(inputImg, outputImg, gdalformat='KEA', noDataVal=0, 
                            clusterer=MiniBatchKMeans(n_clusters=60, init='k-means++', max_iter=100, batch_size=100), 
                            calcStats=True, useMeanShiftEstBandWidth=False, tileXSize=200, tileYSize=200):
    """
A function which allows a clustering to be performed using the algorithms available
within the scikit-learn library. The clusterer is applied to a single tile at a time
and therefore produces tile boundaries in the result. However, memory is controlled
such that usage isn't excessive which it could be when processing a whole image.

:param inputImg: input image file.
:param outputImg: output image file.
:param gdalformat: output image file format.
:param noDataVal: no data value associated with the input image.
:param clusterer: clusterer from scikit-learn which must have a predict function.
:param calcStats: calculate image pixel statistics, histogram and image pyramids - note if you are not using a KEA file then the format needs to support RATs for this option as histogram and colour table are written to RAT.
:param useMeanShiftEstBandWidth: use the mean-shift algorithm as the clusterer (pass None as the clusterer) where the bandwidth is calculated from the data itself.
:param tileXSize: tile size in the x-axis in pixels.
:param tileYSize: tile size in the y-axis in pixels.

"""
    if useMeanShiftEstBandWidth:
        from sklearn.cluster import MeanShift, estimate_bandwidth

    reader = ImageReader(inputImg, windowxsize=tileXSize, windowysize=tileYSize)
    writer = None
    for (info, block) in tqdm.tqdm(reader):
        blkShape = block.shape
        blkBands = block.reshape((blkShape[0], (blkShape[1 ] *blkShape[2]))).T
        ID = numpy.arange(blkBands.shape[0])
        outClusterVals = numpy.zeros((blkBands.shape[0]))

        finiteMskArr = numpy.isfinite(blkBands).all(axis=1)
        ID = ID[finiteMskArr]
        blkBands = blkBands[finiteMskArr]

        noDataValArr = numpy.logical_not(numpy.where(blkBands == noDataVal, True, False).all(axis=1))

        blkBandsNoData = blkBands[noDataValArr]
        ID = ID[noDataValArr]

        if ID.shape[0] > 0:
            if useMeanShiftEstBandWidth:
                bandwidth = estimate_bandwidth(blkBandsNoData, quantile=0.2, n_samples=1000)
                clusterer = MeanShift(bandwidth=bandwidth, bin_seeding=True)

            clusterer.fit(blkBandsNoData)
            outPred = clusterer.labels_ + 1
            outClusterVals[ID] = outPred

        outClusterValsOutArr = outClusterVals.reshape([1 ,blkShape[1] ,blkShape[2]])

        if writer is None:
            writer = ImageWriter(outputImg, info=info, firstblock=outClusterValsOutArr, drivername=gdalformat, creationoptions=[])
        else:
            writer.write(outClusterValsOutArr)
    writer.close(calcStats=False)

    if calcStats:
        rsgislib.rastergis.populateStats(clumps=outputImg, addclrtab=True, calcpyramids=True, ignorezero=True)



def img_pixel_cluster(inputImg, outputImg, gdalformat='KEA', noDataVal=0, 
                      clusterer=MiniBatchKMeans(n_clusters=60, init='k-means++', max_iter=100, batch_size=100), 
                      calcStats=True, useMeanShiftEstBandWidth=False):
    """
A function which allows a clustering to be performed using the algorithms available
within the scikit-learn library. The clusterer is applied to the whole image in one
operation so therefore requires the whole image to be loaded into memory. However,
if there is sufficent memory all the clustering algorithms within scikit-learn can be
applied without boundary artifacts.

:param inputImg: input image file.
:param outputImg: output image file.
:param gdalformat: output image file format.
:param noDataVal: no data value associated with the input image.
:param clusterer: clusterer from scikit-learn which must have a predict function.
:param calcStats: calculate image pixel statistics, histogram and image pyramids - note if you are not using a KEA file then the format needs to support RATs for this option as histogram and colour table are written to RAT.
:param useMeanShiftEstBandWidth: use the mean-shift algorithm as the clusterer (pass None as the clusterer) where the bandwidth is calculated from the data itself.

"""
    # Create output image
    rsgislib.imageutils.createCopyImage(inputImg, outputImg, 1, 0, gdalformat, rsgislib.TYPE_16UINT)

    if useMeanShiftEstBandWidth:
        from sklearn.cluster import MeanShift, estimate_bandwidth

    gdalDS = gdal.Open(inputImg, gdal.GA_ReadOnly)
    nPxls = gdalDS.RasterXSize * gdalDS.RasterYSize

    pxlVals = numpy.zeros((gdalDS.RasterCount, nPxls))

    for nBand in numpy.arange(gdalDS.RasterCount):
        gdalBand = gdalDS.GetRasterBand(int(nBand +1))
        imgArr = gdalBand.ReadAsArray().flatten()
        pxlVals[nBand] = imgArr

    pxlVals = pxlVals.T

    ID = numpy.arange(pxlVals.shape[0])
    outClusterVals = numpy.zeros((pxlVals.shape[0]))

    finiteMskArr = numpy.isfinite(pxlVals).all(axis=1)
    ID = ID[finiteMskArr]
    pxlVals = pxlVals[finiteMskArr]

    noDataValArr = numpy.logical_not(numpy.where(pxlVals == noDataVal, True, False).all(axis=1))

    pxlVals = pxlVals[noDataValArr]
    ID = ID[noDataValArr]

    if ID.shape[0] > 0:
        if useMeanShiftEstBandWidth:
            bandwidth = estimate_bandwidth(pxlVals, quantile=0.2, n_samples=1000)
            clusterer = MeanShift(bandwidth=bandwidth, bin_seeding=True)
        print('Perform Clustering')
        clusterer.fit(pxlVals)
        print('Performed Clustering')
        outPred = clusterer.labels_ + 1
        outClusterVals[ID] = outPred

    outClusterValsOutArr = outClusterVals.reshape([gdalDS.RasterYSize, gdalDS.RasterXSize])
    print(outClusterValsOutArr.shape)

    gdalOutDS = gdal.Open(outputImg, gdal.GA_Update)
    gdalOutBand = gdalOutDS.GetRasterBand(1)
    gdalOutBand.WriteArray(outClusterValsOutArr)
    gdalOutDS = None
    gdalDS = None

    if calcStats:
        rsgislib.rastergis.populateStats(clumps=outputImg, addclrtab=True, calcpyramids=True, ignorezero=True)






#!/usr/bin/env python
"""
The classification module provides classification functionality within RSGISLib.
"""

# import the C++ extension into this level
from ._classification import *

class ClassSimpleInfoObj(object):
    """
    This is a class to store the information associated within the classification.

    :param id: Output pixel value for this class
    :param fileH5: hdf5 file (from rsgislib.imageutils.extractZoneImageBandValues2HDF) with the training
                   data for the class
    :param red: Red colour for visualisation (0-255)
    :param green: Green colour for visualisation (0-255)
    :param blue: Blue colour for visualisation (0-255)

    """

    def __init__(self, id=None, fileH5=None, red=None, green=None, blue=None):
        """

        :param id: Output pixel value for this class
        :param fileH5: hdf5 file (from rsgislib.imageutils.extractZoneImageBandValues2HDF) with the training
               data for the class
        :param red: Red colour for visualisation (0-255)
        :param green: Green colour for visualisation (0-255)
        :param blue: Blue colour for visualisation (0-255)

        """
        self.id = id
        self.fileH5 = fileH5
        self.red = red
        self.green = green
        self.blue = blue


class ClassInfoObj(object):
    """
    This is a class to store the information associated within the classification.

    :param id: Internal unique ID value for this class (must start 0 and be consecutive between the classes)
    :param out_id: External unique ID for ther class which will be used as the output image pixel value, can
                   be any integer.
    :param trainfileH5: hdf5 file (from rsgislib.imageutils.extractZoneImageBandValues2HDF) with the training
                        data for the class
    :param testfileH5: hdf5 file (from rsgislib.imageutils.extractZoneImageBandValues2HDF) with the testing
                       data for the class
    :param validfileH5: hdf5 file (from rsgislib.imageutils.extractZoneImageBandValues2HDF) with the validation
                        data for the class
    :param red: Red colour for visualisation (0-255)
    :param green: Green colour for visualisation (0-255)
    :param blue: Blue colour for visualisation (0-255)

    """

    def __init__(self, id=None, out_id=None, trainfileH5=None, testfileH5=None, validfileH5=None, red=None, green=None,
                 blue=None):
        """

        :param id: Internal unique ID value for this class (must start 0 and be consecutive between the classes)
        :param out_id: External unique ID for ther class which will be used as the output image pixel value,
                       can be any integer.
        :param trainfileH5: hdf5 file (from rsgislib.imageutils.extractZoneImageBandValues2HDF) with the training
                            data for the class
        :param testfileH5: hdf5 file (from rsgislib.imageutils.extractZoneImageBandValues2HDF) with the testing
                           data for the class
        :param validfileH5: hdf5 file (from rsgislib.imageutils.extractZoneImageBandValues2HDF) with the validation
                            data for the class
        :param red: Red colour for visualisation (0-255)
        :param green: Green colour for visualisation (0-255)
        :param blue: Blue colour for visualisation (0-255)

        """
        self.id = id
        self.out_id = out_id
        self.trainfileH5 = trainfileH5
        self.testfileH5 = testfileH5
        self.validfileH5 = validfileH5
        self.red = red
        self.green = green
        self.blue = blue

class ClassVecSamplesInfoObj(object):
    """
    This is a class to store the information associated with the classification vector training regions.

    :param id: Unique ID for the class (will probably be the pixel value for this class)
    :param classname: Unique name for the class.
    :param vecfile: A vector file path with the training samples
    :param veclyr: The vector layer name within the vecfile for the training samples.
    :param fileH5: A file path for a HDF5 file where the pixel values for these samples will be stored.
    """

    def __init__(self, id=None, classname=None,  vecfile=None, veclyr=None, fileH5=None):
        """

        :param id: Unique ID for the class (will probably be the pixel value for this class)
        :param classname: Unique name for the class.
        :param vecfile: A vector file path with the training samples
        :param veclyr: The vector layer name within the vecfile for the training samples.
        :param fileH5: A file path for a HDF5 file where the pixel values for these samples will be stored.

        """
        self.id = id
        self.classname = classname
        self.vecfile = vecfile
        self.veclyr = veclyr
        self.fileH5 = fileH5


class SamplesInfoObj(object):
    """
    This is a class to store the information associated within the classification.

    :param className: The name of the class
    :param classID: Is the classification numeric ID (i.e., output pixel value)
    :param maskImg: The input image mask from which samples are taken
    :param maskPxlVal: The pixel value within the mask for the class
    :param outSampImgFile: Temporary file which will store the sampled pixels.
    :param numSamps: The number of samples required.
    :param samplesH5File: File location for the HDF5 file with the input image values for training.
    :param red: for visualisation red value.
    :param green: for visualisation green value.
    :param blue: for visualisation blue value.

    """

    def __init__(self, className=None, classID=None, maskImg=None, maskPxlVal=None, outSampImgFile=None, numSamps=None,
                 samplesH5File=None, red=None, green=None, blue=None):
        """

        :param className: The name of the class
        :param classID: Is the classification numeric ID (i.e., output pixel value)
        :param maskImg: The input image mask from which samples are taken
        :param maskPxlVal: The pixel value within the mask for the class
        :param outSampImgFile: Temporary file which will store the sampled pixels.
        :param numSamps: The number of samples required.
        :param samplesH5File: File location for the HDF5 file with the input image values for training.
        :param red: for visualisation red value.
        :param green: for visualisation green value.
        :param blue: for visualisation blue value.

        """
        self.className = className
        self.classID = classID
        self.maskImg = maskImg
        self.maskPxlVal = maskPxlVal
        self.outSampImgFile = outSampImgFile
        self.numSamps = numSamps
        self.samplesH5File = samplesH5File
        self.red = red
        self.green = green
        self.blue = blue


def generateTransectAccuracyPts(inputImage, inputLinesShp, outputPtsShp, classImgCol, classImgVecCol,
                                classRefVecCol, lineStep, force=False):
    """ A tool for converting a set of lines in to point transects and 
populating with the information for undertaking an accuracy assessment.

Where:

:param inputImage: is a string specifying the input image file with classification.
:param inputLinesShp: is a string specifying the input lines shapefile path.
:param outputPtsShp: is a string specifying the output points shapefile path.
:param classImgCol: is a string speciyfing the name of the column in the image file containing the class names.
:param classImgVecCol: is a string specifiying the output column in the shapefile for the classified class names.
:param classRefVecCol: is an optional string specifiying an output column in the shapefile which can be used in the
                       accuracy assessment for the reference data.
:param lineStep: is a double specifying the step along the lines between the points
:param force: is an optional boolean specifying whether the output shapefile should be deleted if is already exists
              (True and it will be deleted; Default is False)

    """
    # Import the RSGISLib Image Utils module
    import rsgislib.vectorutils
    rsgislib.vectorutils.createLinesOfPoints(inputLinesShp, outputPtsShp, lineStep, force)
    popClassInfoAccuracyPts(inputImage, outputPtsShp, classImgCol, classImgVecCol, classRefVecCol)


def get_class_training_data(imgBandInfo, classVecSampleInfo, tmpdir, sub_sample=None, refImg=None):
    """
    A function to extract training for vector regions for a given input image set.

    :param imgBandInfo: A list of rsgislib.imageutils.ImageBandInfo objects to define the images and bands of interest.
    :param classVecSampleInfo: A list of rsgislib.classification.ClassVecSamplesInfoObj objects to define the
                               training regions.
    :param tmpdir: A directory for temporary outputs created during the processing.
    :param sub_sample: If not None then an integer needs to be provided which takes a random selection from the
                       available samples to balance the number of samples used for the classification.
    :param refImg: A reference image which defines the area of interest, pixel size etc. for the processing.
                   If None then an image will be generated using the input images but the tmpdir needs to be defined.
    :return: dictionary of ClassSimpleInfoObj objects.

    """
    import rsgislib
    import rsgislib.imageutils
    import rsgislib.vectorutils
    import os
    import os.path
    import random
    import shutil

    # Get valid mask, rasterised to this
    rsgis_utils = rsgislib.RSGISPyUtils()
    uid_str = rsgis_utils.uidGenerator()
    tmp_lcl_dir = os.path.join(tmpdir, "get_class_training_data_{}".format(uid_str))
    if not os.path.exists(tmp_lcl_dir):
        os.makedirs(tmp_lcl_dir)

    rasterise_ref_img = refImg
    if refImg is None:
        rasterise_ref_img = os.path.join(tmp_lcl_dir, "ref_img_vmsk.kea")
        rsgislib.imageutils.create_valid_mask(imgBandInfo, rasterise_ref_img, 'KEA', tmp_lcl_dir)

    classInfo = dict()
    for class_sample_info in classVecSampleInfo:
        cls_basename = rsgis_utils.get_file_basename(class_sample_info.fileH5)
        out_vec_img = os.path.join(tmp_lcl_dir, "{}_img.kea".format(cls_basename))
        rsgislib.vectorutils.rasteriseVecLyr(class_sample_info.vecfile, class_sample_info.veclyr, rasterise_ref_img,
                                             out_vec_img, gdalformat="KEA", burnVal=class_sample_info.id,
                                             datatype=rsgislib.TYPE_16UINT, vecAtt=None, vecExt=False, thematic=True,
                                             nodata=0)

        if sub_sample is not None:
            out_vec_img_subsample = os.path.join(tmp_lcl_dir, "{}_img_subsample.kea".format(cls_basename))
            rsgislib.imageutils.performRandomPxlSampleInMaskLowPxlCount(out_vec_img, out_vec_img_subsample,
                                                                        'KEA', maskvals=[class_sample_info.id],
                                                                        numSamples=sub_sample)
            out_vec_img = out_vec_img_subsample

        rsgislib.imageutils.extractZoneImageBandValues2HDF(imgBandInfo, out_vec_img, class_sample_info.fileH5,
                                                           class_sample_info.id)
        rand_red_val = random.randint(1, 255)
        rand_grn_val = random.randint(1, 255)
        rand_blu_val = random.randint(1, 255)
        classInfo[class_sample_info.classname] = ClassSimpleInfoObj(id=class_sample_info.id,
                                                                    fileH5=class_sample_info.fileH5,
                                                                    red=rand_red_val,
                                                                    green=rand_grn_val,
                                                                    blue=rand_blu_val)

    shutil.rmtree(tmp_lcl_dir)
    return classInfo


def get_class_training_chips_data(imgBandInfo, classVecSampleInfo, chip_h_size, tmpdir, refImg=None):
    """
    A function to extract training chips (windows/regions) for vector regions for a given input image set.

    :param imgBandInfo: A list of rsgislib.imageutils.ImageBandInfo objects to define the images and bands of interest.
    :param classVecSampleInfo: A list of rsgislib.classification.ClassVecSamplesInfoObj objects to define the
                               training regions.
    :param chip_h_size: is half the chip size to be extracted (i.e., 10 with output image chips 21x21, 10 pixels
                        either size of the one of interest).
    :param tmpdir: A directory for temporary outputs created during the processing.
    :param refImg: A reference image which defines the area of interest, pixel size etc. for the processing.
                   If None then an image will be generated using the input images but the tmpdir needs to be defined.
    :return: dictionary of ClassSimpleInfoObj objects.

    """
    import rsgislib
    import rsgislib.imageutils
    import rsgislib.vectorutils
    import os
    import os.path
    import random
    import shutil

    # Get valid mask, rasterised to this
    rsgis_utils = rsgislib.RSGISPyUtils()
    uid_str = rsgis_utils.uidGenerator()
    tmp_lcl_dir = os.path.join(tmpdir, "get_class_training_chips_data_{}".format(uid_str))
    if not os.path.exists(tmp_lcl_dir):
        os.makedirs(tmp_lcl_dir)

    rasterise_ref_img = refImg
    if refImg is None:
        rasterise_ref_img = os.path.join(tmp_lcl_dir, "ref_img_vmsk.kea")
        rsgislib.imageutils.create_valid_mask(imgBandInfo, rasterise_ref_img, 'KEA', tmp_lcl_dir)

    classInfo = dict()
    for class_sample_info in classVecSampleInfo:
        cls_basename = rsgis_utils.get_file_basename(class_sample_info.fileH5)
        out_vec_img = os.path.join(tmp_lcl_dir, "{}_img.kea".format(cls_basename))
        rsgislib.vectorutils.rasteriseVecLyr(class_sample_info.vecfile, class_sample_info.veclyr, rasterise_ref_img,
                                             out_vec_img, gdalformat="KEA", burnVal=class_sample_info.id,
                                             datatype=rsgislib.TYPE_16UINT, vecAtt=None, vecExt=False, thematic=True,
                                             nodata=0)
        rsgislib.imageutils.extractChipZoneImageBandValues2HDF(imgBandInfo, out_vec_img, class_sample_info.id,
                                                               chip_h_size, class_sample_info.fileH5)
        rand_red_val = random.randint(1, 255)
        rand_grn_val = random.randint(1, 255)
        rand_blu_val = random.randint(1, 255)
        classInfo[class_sample_info.classname] = ClassSimpleInfoObj(id=class_sample_info.id,
                                                                    fileH5=class_sample_info.fileH5, red=rand_red_val,
                                                                    green=rand_grn_val, blue=rand_blu_val)
    shutil.rmtree(tmp_lcl_dir)
    return classInfo


def get_num_samples(input_sample_h5_file):
    """
    A function to return the number of samples within the input HDF5 file.

    :param input_sample_h5_file: Input HDF file, probably from rsgislib.imageutils.extractZoneImageBandValues2HDF.
    :return: the number of samples in the hdf5 file.
    """
    import h5py
    f = h5py.File(input_sample_h5_file, 'r')
    n_samples = f['DATA/DATA'].shape[0]
    f.close()
    return n_samples


def split_sample_train_valid_test(input_sample_h5_file, train_h5_file, valid_h5_file, test_h5_file, test_sample,
                                  valid_sample, train_sample=None, rand_seed=42, datatype=None):
    """
    A function to split a HDF5 samples file (from rsgislib.imageutils.extractZoneImageBandValues2HDF)
    into three (i.e., Training, Validation and Testing).

    :param input_sample_h5_file: Input HDF file, probably from rsgislib.imageutils.extractZoneImageBandValues2HDF.
    :param train_h5_file: Output file with the training data samples (this has the number of samples left following
                          the removal of the test and valid samples if train_sample=None)
    :param valid_h5_file: Output file with the valid data samples.
    :param test_h5_file: Output file with the testing data samples.
    :param test_sample: The size of the testing sample to be taken.
    :param valid_sample: The size of the validation sample to be taken.
    :param train_sample: The size of the training sample to be taken. If None then the remaining samples are returned.
    :param rand_seed: The random seed to be used to randomly select the sub-samples.
    :param datatype: is the data type used for the output HDF5 file (e.g., rsgislib.TYPE_32FLOAT). If None (default)
                     then the output data type will be float32.

    """
    import rsgislib
    import rsgislib.imageutils
    import os
    import os.path
    rsgis_utils = rsgislib.RSGISPyUtils()
    uid_str = rsgis_utils.uidGenerator()
    out_dir = os.path.split(os.path.abspath(test_h5_file))[0]
    tmp_train_valid_sample_file = os.path.join(out_dir, "train_valid_tmp_sample_{}.h5".format(uid_str))
    rsgislib.imageutils.splitSampleHDF5File(input_sample_h5_file, test_h5_file, tmp_train_valid_sample_file,
                                            test_sample, rand_seed, datatype)
    if train_sample is not None:
        tmp_train_sample_file = os.path.join(out_dir, "train_tmp_sample_{}.h5".format(uid_str))
        rsgislib.imageutils.splitSampleHDF5File(tmp_train_valid_sample_file, valid_h5_file, tmp_train_sample_file,
                                                valid_sample, rand_seed, datatype)
        tmp_remain_sample_file = os.path.join(out_dir, "remain_tmp_sample_{}.h5".format(uid_str))
        rsgislib.imageutils.splitSampleHDF5File(tmp_train_sample_file, train_h5_file, tmp_remain_sample_file,
                                                train_sample, rand_seed, datatype)
        os.remove(tmp_train_sample_file)
        os.remove(tmp_remain_sample_file)
    else:
        rsgislib.imageutils.splitSampleHDF5File(tmp_train_valid_sample_file, valid_h5_file, train_h5_file, valid_sample,
                                                rand_seed, datatype)
    os.remove(tmp_train_valid_sample_file)


def split_chip_sample_train_valid_test(input_sample_h5_file, train_h5_file, valid_h5_file, test_h5_file,
                                       test_sample, valid_sample, train_sample=None, rand_seed=42, datatype=None):
    """
    A function to split a chip HDF5 samples file (from rsgislib.imageutils.extractChipZoneImageBandValues2HDF)
    into three (i.e., Training, Validation and Testing).

    :param input_sample_h5_file: Input HDF file, probably from rsgislib.imageutils.extractZoneImageBandValues2HDF.
    :param train_h5_file: Output file with the training data samples (this has the number of samples left following
                          the removal of the test and valid samples if train_sample=None)
    :param valid_h5_file: Output file with the valid data samples.
    :param test_h5_file: Output file with the testing data samples.
    :param test_sample: The size of the testing sample to be taken.
    :param valid_sample: The size of the validation sample to be taken.
    :param train_sample: The size of the training sample to be taken. If None then the remaining samples are returned.
    :param rand_seed: The random seed to be used to randomly select the sub-samples.
    :param datatype: is the data type used for the output HDF5 file (e.g., rsgislib.TYPE_32FLOAT). If None (default)
                     then the output data type will be float32.

    """
    import rsgislib
    import rsgislib.imageutils
    import os
    import os.path

    rsgis_utils = rsgislib.RSGISPyUtils()
    uid_str = rsgis_utils.uidGenerator()
    out_dir = os.path.split(os.path.abspath(test_h5_file))[0]
    tmp_train_valid_sample_file = os.path.join(out_dir, "train_valid_tmp_sample_{}.h5".format(uid_str))
    rsgislib.imageutils.splitSampleChipHDF5File(input_sample_h5_file, test_h5_file, tmp_train_valid_sample_file,
                                                test_sample, rand_seed, datatype)
    if train_sample is not None:
        tmp_train_sample_file = os.path.join(out_dir, "train_tmp_sample_{}.h5".format(uid_str))
        rsgislib.imageutils.splitSampleChipHDF5File(tmp_train_valid_sample_file, valid_h5_file, tmp_train_sample_file,
                                                    valid_sample, rand_seed, datatype)
        tmp_remain_sample_file = os.path.join(out_dir, "remain_tmp_sample_{}.h5".format(uid_str))
        rsgislib.imageutils.splitSampleChipHDF5File(tmp_train_sample_file, train_h5_file, tmp_remain_sample_file,
                                                    train_sample, rand_seed, datatype)
        os.remove(tmp_train_sample_file)
        os.remove(tmp_remain_sample_file)
    else:
        rsgislib.imageutils.splitSampleChipHDF5File(tmp_train_valid_sample_file, valid_h5_file, train_h5_file,
                                                    valid_sample, rand_seed, datatype)
    os.remove(tmp_train_valid_sample_file)


def split_chip_sample_ref_train_valid_test(input_sample_h5_file, train_h5_file, valid_h5_file, test_h5_file,
                                           test_sample, valid_sample, train_sample=None, rand_seed=42, datatype=None):
    """
    A function to split a chip HDF5 samples file (from rsgislib.imageutils.extractChipZoneImageBandValues2HDF)
    into three (i.e., Training, Validation and Testing).

    :param input_sample_h5_file: Input HDF file, probably from rsgislib.imageutils.extractZoneImageBandValues2HDF.
    :param train_h5_file: Output file with the training data samples (this has the number of samples left following
                          the removal of the test and valid samples if train_sample=None)
    :param valid_h5_file: Output file with the valid data samples.
    :param test_h5_file: Output file with the testing data samples.
    :param test_sample: The size of the testing sample to be taken.
    :param valid_sample: The size of the validation sample to be taken.
    :param train_sample: The size of the training sample to be taken. If None then the remaining samples are returned.
    :param rand_seed: The random seed to be used to randomly select the sub-samples.
    :param datatype: is the data type used for the output HDF5 file (e.g., rsgislib.TYPE_32FLOAT). If None (default)
                     then the output data type will be float32.

    """
    import rsgislib
    from rsgislib.imageutils import splitSampleRefChipHDF5File
    import os
    import os.path

    rsgis_utils = rsgislib.RSGISPyUtils()
    uid_str = rsgis_utils.uidGenerator()
    out_dir = os.path.split(os.path.abspath(test_h5_file))[0]
    tmp_train_valid_sample_file = os.path.join(out_dir, "train_valid_tmp_sample_{}.h5".format(uid_str))
    splitSampleRefChipHDF5File(input_sample_h5_file, test_h5_file, tmp_train_valid_sample_file,
                               test_sample, rand_seed, datatype)
    if train_sample is not None:
        tmp_train_sample_file = os.path.join(out_dir, "train_tmp_sample_{}.h5".format(uid_str))
        splitSampleRefChipHDF5File(tmp_train_valid_sample_file, valid_h5_file, tmp_train_sample_file,
                                   valid_sample, rand_seed, datatype)
        tmp_remain_sample_file = os.path.join(out_dir, "remain_tmp_sample_{}.h5".format(uid_str))
        splitSampleRefChipHDF5File(tmp_train_sample_file, train_h5_file, tmp_remain_sample_file,
                                   train_sample, rand_seed, datatype)
        os.remove(tmp_train_sample_file)
        os.remove(tmp_remain_sample_file)
    else:
        splitSampleRefChipHDF5File(tmp_train_valid_sample_file, valid_h5_file, train_h5_file,
                                   valid_sample, rand_seed, datatype)
    os.remove(tmp_train_valid_sample_file)


def flipChipHDF5File(input_h5_file, output_h5_file, datatype=None):
    """
    A function which flips each sample in both the x and y axis. So the
    output file will have double the number of samples as the input file.

    :param input_h5_file: The input HDF5 file for chips extracted from images.
    :param output_h5_file: The output HDF5 file for chips extracted from images.
    :param datatype: is the data type used for the output HDF5 file (e.g., rsgislib.TYPE_32FLOAT). If None (default)
                     then the output data type will be float32.

    """
    import tqdm
    import h5py
    import numpy
    import rsgislib
    rsgis_utils = rsgislib.RSGISPyUtils()
    if datatype is None:
        datatype = rsgislib.TYPE_32FLOAT

    f = h5py.File(input_h5_file, 'r')
    n_in_feats = f['DATA/DATA'].shape[0]
    chip_size = f['DATA/DATA'].shape[1]
    n_bands = f['DATA/DATA'].shape[3]

    n_out_feats = n_in_feats * 2

    feat_arr = numpy.zeros([n_out_feats, chip_size, chip_size, n_bands], dtype=numpy.float32)

    i_feat = 0
    for n in tqdm.tqdm(range(n_in_feats)):
        numpy.copyto(feat_arr[i_feat], numpy.flip(f['DATA/DATA'][n].T, axis=1).T, casting='safe')
        i_feat += 1
        numpy.copyto(feat_arr[i_feat], numpy.flip(f['DATA/DATA'][n].T, axis=2).T, casting='safe')
        i_feat += 1

    f.close()

    ######################################################################
    # Create the output HDF5 file and populate with data.
    ######################################################################
    h5_dtype = rsgis_utils.getNumpyCharCodesDataType(datatype)
    fH5Out = h5py.File(output_h5_file, 'w')
    dataGrp = fH5Out.create_group("DATA")
    metaGrp = fH5Out.create_group("META-DATA")
    dataGrp.create_dataset('DATA', data=feat_arr, chunks=(250, chip_size, chip_size, n_bands),
                           compression="gzip", shuffle=True, dtype=h5_dtype)
    describDS = metaGrp.create_dataset("DESCRIPTION", (1,), dtype="S10")
    describDS[0] = 'IMAGE REF TILES'.encode()
    fH5Out.close()
    ######################################################################


def flipRefChipHDF5File(input_h5_file, output_h5_file, datatype=None):
    """
    A function which flips each sample in both the x and y axis. So the
    output file will have double the number of samples as the input file.

    :param input_h5_file: The input HDF5 file for chips extracted from images.
    :param output_h5_file: The output HDF5 file for chips extracted from images.
    :param datatype: is the data type used for the output HDF5 file (e.g., rsgislib.TYPE_32FLOAT). If None (default)
                     then the output data type will be float32.

    """
    import tqdm
    import h5py
    import numpy
    import rsgislib
    rsgis_utils = rsgislib.RSGISPyUtils()
    if datatype is None:
        datatype = rsgislib.TYPE_32FLOAT

    f = h5py.File(input_h5_file, 'r')
    n_in_feats = f['DATA/REF'].shape[0]
    chip_size = f['DATA/REF'].shape[1]
    n_bands = f['DATA/DATA'].shape[3]

    n_out_feats = n_in_feats * 2

    feat_arr = numpy.zeros([n_out_feats, chip_size, chip_size, n_bands], dtype=numpy.float32)
    feat_ref_arr = numpy.zeros([n_out_feats, chip_size, chip_size], dtype=numpy.uint16)

    i_feat = 0
    for n in tqdm.tqdm(range(n_in_feats)):
        numpy.copyto(feat_ref_arr[i_feat], numpy.flip(f['DATA/REF'][n].T, axis=0).T, casting='safe')
        numpy.copyto(feat_arr[i_feat], numpy.flip(f['DATA/DATA'][n].T, axis=1).T, casting='safe')
        i_feat += 1
        numpy.copyto(feat_ref_arr[i_feat], numpy.flip(f['DATA/REF'][n].T, axis=1).T, casting='safe')
        numpy.copyto(feat_arr[i_feat], numpy.flip(f['DATA/DATA'][n].T, axis=2).T, casting='safe')
        i_feat += 1

    f.close()

    ######################################################################
    # Create the output HDF5 file and populate with data.
    ######################################################################
    h5_dtype = rsgis_utils.getNumpyCharCodesDataType(datatype)
    fH5Out = h5py.File(output_h5_file, 'w')
    dataGrp = fH5Out.create_group("DATA")
    metaGrp = fH5Out.create_group("META-DATA")
    dataGrp.create_dataset('DATA', data=feat_arr, chunks=(250, chip_size, chip_size, n_bands),
                           compression="gzip", shuffle=True, dtype=h5_dtype)
    dataGrp.create_dataset('REF', data=feat_ref_arr, chunks=(250, chip_size, chip_size),
                           compression="gzip", shuffle=True, dtype='H')
    describDS = metaGrp.create_dataset("DESCRIPTION", (1,), dtype="S10")
    describDS[0] = 'IMAGE REF TILES'.encode()
    fH5Out.close()
    ######################################################################


def label_pxl_sample_chips(sample_pxls_img, cls_msk_img, output_image, gdalformat, chip_size, cls_lut,
                           sample_pxl_img_band=1, cls_msk_img_band=1):
    """
    A function which labels image pixels based on the proportions of a class within a chip around the
    pixel (can be used in combination with rsgislib.imageutils.assign_random_pxls). It is expected that
    this function will be used when trying to use existing maps to create deep learning chip classification
    training data.

    Pixels are labelled if the proportion of pixels is >= the threshold provided in the LUT. If more than
    one class meets the threshold then the one with the highest proportion is assigned.

    :param sample_pxls_img: The input binary image with the pixel locations (value == 1)
    :param cls_msk_img: The classification image used to assign the output pixel values.
    :param output_image: The output image. Single pixels with the class value will be outputted.
    :param gdalformat: The output image file format.
    :param chip_size: The size of the chip used to identify the class - would probably correspond
                      to the chip size being used for the deep learning classification. Areas used
                      is half the chip size around the pixel (i.e., the pixel from the samples image
                      will be at the centre of the chip).
    :param cls_lut: A dict look up table (LUT) with the thresholds per class for the pixel to be
                    classified as that class.
    :param sample_pxl_img_band: Default 1. The image band in the sample image.
    :param cls_msk_img_band: Default 1. The image band in the sample image.

    Example::

        sample_pxls_img = 'LS5TM_20000108_latn531lonw37_r23p204_osgb_samples.kea'
        cls_msk_img = 'LS5TM_20000108_latn531lonw37_r23p204_osgb_clouds_up.kea'
        output_image = 'LS5TM_20000108_latn531lonw37_r23p204_osgb_samples_lbld.kea'

        cls_lut = dict()
        cls_lut[1] = 0.2
        cls_lut[2] = 0.2
        cls_lut[3] = 0.99

        label_pxl_sample_chips(sample_pxls_img, cls_msk_img, output_image, 'KEA', 21, cls_lut)

    """
    import rsgislib.rastergis
    from rios.imagereader import ImageReader
    from rios.imagewriter import ImageWriter
    import tqdm
    import numpy
    import math

    chip_size_odd = False
    if (chip_size % 2) != 0:
        chip_size_odd = True

    img_win_h_size = math.floor(chip_size / 2)
    img_win_size = chip_size
    n_pxls = img_win_size * img_win_size

    inImgs = list()
    inImgBands = list()
    inImgs.append(sample_pxls_img)
    inImgBands.append([sample_pxl_img_band])
    inImgs.append(cls_msk_img)
    inImgBands.append([cls_msk_img_band])

    writer = None
    reader = ImageReader(inImgs, windowxsize=200, windowysize=200, overlap=img_win_h_size, layerselection=inImgBands)
    for (info, block) in tqdm.tqdm(reader):
        samples_msk_arr = block[0]
        blk_shp = samples_msk_arr.shape

        xSize = blk_shp[2] - (img_win_h_size * 2)
        ySize = blk_shp[1] - (img_win_h_size * 2)
        xRange = numpy.arange(img_win_h_size, img_win_h_size + xSize, 1)
        yRange = numpy.arange(img_win_h_size, img_win_h_size + ySize, 1)
        out_samp_arr = numpy.zeros_like(samples_msk_arr, dtype=numpy.uint8)
        for y in yRange:
            yMin = y - img_win_h_size
            yMax = y + img_win_h_size
            if chip_size_odd:
                yMax += 1
            for x in xRange:
                xMin = x - img_win_h_size
                xMax = x + img_win_h_size
                if chip_size_odd:
                    xMax += 1
                if samples_msk_arr[0][y][x] == 1:
                    img_blk = block[1][0, yMin:yMax, xMin:xMax]
                    uniq_vals, uniq_counts = numpy.unique(img_blk, return_counts=True)
                    uniq_dict = dict(zip(uniq_vals, uniq_counts))
                    first = True
                    for val in uniq_vals:
                        if val in cls_lut:
                            val_prop = uniq_dict[val] / n_pxls
                            if val_prop >= cls_lut[val]:
                                if first:
                                    max_val = val
                                    max_val_prop = val_prop
                                    first = False
                                elif val_prop > max_val_prop:
                                    max_val = val
                                    max_val_prop = val_prop
                    if not first:
                        out_samp_arr[0][y][x] = max_val

        if writer is None:
            writer = ImageWriter(output_image, info=info, firstblock=out_samp_arr, drivername=gdalformat)
        else:
            writer.write(out_samp_arr)
    writer.close(calcStats=False)

    rsgislib.rastergis.populateStats(output_image, True, True, True)


#!/usr/bin/env python
"""
These utilities allow for easier use of the Look Up Table (LUT) functionality within
RSGISLib.
"""
############################################################################
#  imagelut.py
#
#  Copyright 2019 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide a set of utilities for working with image LUTs in RSGISLib
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 23/10/2019
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
############################################################################

import rsgislib
import os.path

haveGDALPy = True
try:
    import osgeo.gdal as gdal, ogr
except ImportError as gdalErr:
    haveGDALPy = False


def createImgExtentLUT(imgList, vectorFile, vectorLyr, vecDriver, ignore_none_imgs=False, out_proj_wgs84=False,
                       overwrite_lut_file=False):
    """
    Create a vector layer look up table (LUT) for a directory of images.

    This function is the same as rsgislib.vectorutils.createImgExtentLUT and just calls that function but
    is here as it might be a more logical place for people to find this functionality.

    :param imgList: list of input images for the LUT. All input images should be the same projection/coordinate system.
    :param vectorFile: output vector file/path
    :param vectorLyr: output vector layer
    :param vecDriver: the output vector layer type (e.g., GPKG).
    :param ignore_none_imgs: if a NULL epsg is returned from an image then ignore and don't include in LUT else throw exception.
    :param out_proj_wgs84: if True then the image bounding boxes will be re-projected to EPSG:4326.
    :param overwrite_lut_file: if True then output file will be overwritten. If false then not, e.g., can add extra layer to GPKG

    Example::

        import glob
        import rsgislib.imageutils.imagelut
        imgList = glob.glob('/Users/pete/Temp/GabonLandsat/Hansen*.kea')
        rsgislib.imageutils.imagelut.createImgExtentLUT(imgList, './ImgExtents.gpkg', 'HansenImgExtents', 'GPKG')

    """
    import rsgislib.vectorutils
    rsgislib.vectorutils.createImgExtentLUT(imgList, vectorFile, vectorLyr, vecDriver, ignore_none_imgs,
                                            out_proj_wgs84, overwrite_lut_file)


def query_img_lut(scn_bbox, lutdbfile, lyrname):
    """

    :param scn_bbox: A bbox (MinX, MaxX, MinY, MaxY) in the same projection as the LUT for the area of interest.
    :param lutdbfile: The file path to the LUT datbase file (e.g., lut.gpkg).
    :param lyrname: The layer name within the database file.
    :return: a list of files from the LUT
    """
    import rsgislib.vectorutils
    fileListLUT = rsgislib.vectorutils.getAttLstSelectBBoxFeats(lutdbfile, lyrname, ['path', 'filename'], scn_bbox)
    imgs = []
    for item in fileListLUT:
        print(item)
        imgs.append(os.path.join(item['path'], item['filename']))
    return imgs


def get_all_lut_imgs(lutdbfile, lyrname):
    """
    Get a list of all the images within the LUT.
    :param lutdbfile: The file path to the LUT datbase file (e.g., lut.gpkg).
    :param lyrname: The layer name within the database file.
    :return: a list of files from the LUT
    """
    import rsgislib.vectorutils
    fileListLUT = rsgislib.vectorutils.readVecColumns(lutdbfile, lyrname, ['path', 'filename'])
    imgs = []
    for item in fileListLUT:
        print(item)
        imgs.append(os.path.join(item['path'], item['filename']))
    return imgs


def getRasterLyr(scn_bbox, lutdbfile, lyrname, tmp_path):
    """
    This function provides a single raster layer using the LUT file provided. If
    a single image file intersecting with the BBOX is not within the LUT then a
    VRT generated from a number of images will be returned. The VRT will be within
    the tmp_path directory which will need to be cleaned up afterwards. If no image(s)
    is available then None is returned.

    :param scn_bbox: A bbox (MinX, MaxX, MinY, MaxY) in the same projection as the LUT for the area of interest.
    :param lutdbfile: The file path to the LUT datbase file (e.g., lut.gpkg).
    :param lyrname: The layer name within the database file.
    :param tmp_path: A directory where temporary files can be saved to.
    :return: a string with the file path to the raster layer. If no image intersects then None is returned.
    """
    if not haveGDALPy:
        raise Exception("Need to have GDAL library available for getRasterLyr function to work.")
    import rsgislib.vectorutils

    rsgis_utils = rsgislib.RSGISPyUtils()
    fileListLUT = rsgislib.vectorutils.getAttLstSelectBBoxFeats(lutdbfile, lyrname, ['path', 'filename'], scn_bbox)

    imgbase = "imglyr_{}".format(rsgis_utils.uidGenerator())
    # if number of scenes available is > 0 then create VRT
    if len(fileListLUT) > 1:
        imgs = []
        for item in fileListLUT:
            imgs.append(os.path.join(item['path'], item['filename']))
        outimgfile = os.path.join(tmp_path, '{}_tmp.vrt'.format(imgbase))
        gdal.BuildVRT(outimgfile, imgs)
    elif len(fileListLUT) == 1:
        outimgfile = os.path.join(fileListLUT[0]['path'], fileListLUT[0]['filename'])
    else:
        outimgfile = None
    return outimgfile


